"""PathwayCommons module. Depends on HGNC."""

import warnings
from typing import Dict

import pandas as pd
from pyorientdb import OrientDB
from tqdm import tqdm

from ebel.constants import RID
from ebel.manager.orientdb import odb_meta, odb_structure, urls
from ebel.manager.orientdb.biodbs.hgnc import Hgnc
from ebel.manager.orientdb.constants import PATHWAY_COMMONS
from ebel.manager.rdbms.models import hgnc
from ebel.manager.rdbms.models import pathway_commons as pc
from ebel.tools import get_file_path

warnings.filterwarnings("ignore", "This pattern has match groups")


class PathwayCommons(odb_meta.Graph):
    """Pathway Commons."""

    def __init__(self, client: OrientDB = None):
        """Init Pathway Commons."""
        self.client = client
        self.biodb_name = PATHWAY_COMMONS
        self.url = urls.PATHWAY_COMMONS
        self.urls = {self.biodb_name: self.url}
        self.file_path = get_file_path(self.url, self.biodb_name)

        super().__init__(
            generics=odb_structure.pathway_commons_generics,
            edges=odb_structure.pathway_commons_edges,
            tables_base=pc.Base,
            urls=self.urls,
            biodb_name=self.biodb_name,
        )
        self.hgnc = Hgnc(self.client)

    def __len__(self) -> int:
        """Get number of edges in OrientDB."""
        return self.number_of_edges

    def __repr__(self) -> str:
        """Represent PathwayCommons Integration as string."""
        template = "{{BioDatabase:PathwayCommons}}[url:{url}, edges:{edges}, nodes:{generics}]"
        representation = template.format(url=self.url, edges=self.number_of_edges, generics=self.number_of_generics)
        return representation

    def __contains__(self, rs_number: int) -> bool:
        """Checks if RS number (without prefix RS) exists in BEL graph."""
        return self.entry_exists(self.biodb_name, rs_number=rs_number)

    def insert_data(self) -> Dict[str, int]:
        """Insert data in generic OrientDB class."""
        usecols = [
            "PARTICIPANT_A",
            "INTERACTION_TYPE",
            "PARTICIPANT_B",
            "INTERACTION_DATA_SOURCE",
            "INTERACTION_PUBMED_ID",
            "PATHWAY_NAMES",
        ]

        df = pd.read_csv(self.file_path, sep="\t", low_memory=True, usecols=usecols)
        # Because 2 tables are in file, we have to identify where second table starts and slice the dataframe
        df = df.iloc[: df[df["PARTICIPANT_A"] == "PARTICIPANT"].index[0]]

        df.columns = self._standardize_column_names(df.columns)
        df.pathway_names = df.pathway_names.str.split(";")
        df.interaction_data_source = df.interaction_data_source.str.split(";")
        df.interaction_pubmed_id = df.interaction_pubmed_id.str.split(";")

        df.index += 1
        df["id"] = df.index

        # insert main table
        df[["participant_a", "participant_b", "interaction_type"]].to_sql(
            pc.PathwayCommons.__tablename__,
            self.engine,
            if_exists="append",
            index=False,
            chunksize=10000,
        )

        # insert pmids, names, sources
        self.create_pmids_table(df)
        df_pc_names = self.create_names_table(df)
        df_pc_sources = self.create_sources_table(df)

        # joiing tables
        self.create_joining_table_sources(df, df_pc_sources)
        self.create_joining_table_names(df, df_pc_names)

        self.session.commit()
        return {self.biodb_name: df.shape[0]}

    def create_pmids_table(self, df):
        """Create the Pmid table."""
        df_pmids = df[["id", "interaction_pubmed_id"]].dropna().explode("interaction_pubmed_id")
        df_pmids.rename(
            columns={"id": "pathway_commons_id", "interaction_pubmed_id": "pmid"},
            inplace=True,
        )
        df_pmids.pmid = pd.to_numeric(df_pmids.pmid, errors="coerce")
        df_pmids.to_sql(
            pc.Pmid.__tablename__,
            con=self.engine,
            index=False,
            if_exists="append",
            chunksize=10000,
        )
        del df_pmids

    def create_joining_table_names(self, df, df_pc_names):
        """Create the joining table for Names."""
        df_pc_names_pc = df[["pathway_names", "id"]].explode("pathway_names").dropna()
        df_pc_names_pc.rename(columns={"id": "pathway_commons_id", "pathway_names": "name"}, inplace=True)
        df_pc_names_pc.set_index("name", inplace=True)
        df_pc_names["pathway_commons_pathway_name_id"] = df_pc_names.index
        df_pc_names.set_index("name", inplace=True)
        df_pc_names_pc.join(df_pc_names, how="inner").to_sql(
            pc.pathway_commons__pathway_name.fullname,
            con=self.engine,
            index=False,
            if_exists="append",
            chunksize=10000,
        )
        del df_pc_names_pc
        del df_pc_names

    def create_joining_table_sources(self, df, df_pc_sources):
        """Create the joining table for Source."""
        df_pc_sources_pc = df[["interaction_data_source", "id"]].explode("interaction_data_source").dropna()
        df_pc_sources_pc.rename(
            columns={"id": "pathway_commons_id", "interaction_data_source": "source"},
            inplace=True,
        )
        df_pc_sources_pc.set_index("source", inplace=True)
        df_pc_sources["pathway_commons_source_id"] = df_pc_sources.index
        df_pc_sources.set_index("source", inplace=True)
        df_pc_sources_pc.join(df_pc_sources, how="inner").to_sql(
            pc.pathway_commons__source.fullname,
            con=self.engine,
            index=False,
            if_exists="append",
            chunksize=10000,
        )
        del df_pc_sources_pc
        del df_pc_sources

    def create_sources_table(self, df):
        """Create the Sources table."""
        pc_sources = df.interaction_data_source.explode().str.strip().dropna().unique()
        df_pc_sources = pd.DataFrame(pc_sources, columns=["source"])
        df_pc_sources.index += 1
        df_pc_sources.index.rename("id", inplace=True)
        df_pc_sources.to_sql(pc.Source.__tablename__, self.engine, if_exists="append")
        return df_pc_sources

    def create_names_table(self, df):
        """Create the Names table."""
        pc_names = df.pathway_names.explode().str.strip().dropna().unique()
        df_pc_names = pd.DataFrame(pc_names, columns=["name"])
        df_pc_names.index += 1
        df_pc_names.index.rename("id", inplace=True)
        df_pc_names.to_sql(pc.PathwayName.__tablename__, self.engine, if_exists="append")
        return df_pc_names

    def get_pathway_name_rid_dict(self) -> Dict[str, str]:
        """Get dict of pathway names as keys and their rIDs as values."""
        pathway_name_rid_dict = {}
        number = self.get_number_of_class("pc_pathway_name")

        # insert data from rdbms in odb if not exists
        if number == 0:
            for (pc_pathway_name,) in self.session.query(pc.PathwayName.name).all():
                value_dict = {"name": pc_pathway_name}
                rid = self.insert_record("pc_pathway_name", value_dict)
                pathway_name_rid_dict[pc_pathway_name] = rid
        # get data from odb
        else:
            pc_pathway_names = self.query_class(class_name="pc_pathway_name", columns=["name"])
            pathway_name_rid_dict = {x["name"]: x[RID] for x in pc_pathway_names}
        return pathway_name_rid_dict

    def _get_rid(self, rid, valid, name, pure_symbol_rids_dict):
        """Get the rID of an entry based on passed metadata."""
        _rid = None
        if isinstance(rid, str):
            _rid = rid
        else:
            if valid and name:
                if name in pure_symbol_rids_dict:
                    _rid = pure_symbol_rids_dict[name]
                else:
                    _rid = self.insert_record(
                        "protein",
                        {
                            "name": name,
                            "namespace": "HGNC",
                            "bel": f'p(HGNC:"{name}")',
                            "pure": True,
                        },
                    )
                    pure_symbol_rids_dict[name] = _rid
        return _rid

    def update_interactions(self) -> Dict[str, int]:
        """Update pathway commons interactions in graph."""
        inserted = {}

        pc_pathway_name_rid_dict = self.get_pathway_name_rid_dict()
        valid_hgnc_symbols = {x[0] for x in self.session.query(hgnc.Hgnc).with_entities(hgnc.Hgnc.symbol).all()}

        cols = ["symbol", "rid"]
        pure_symbol_rids_dict = self.hgnc.get_pure_symbol_rids_dict()
        df_all = pd.DataFrame(pure_symbol_rids_dict.items(), columns=cols)
        df_bel = pd.DataFrame(self.hgnc.get_pure_symbol_rids_dict_in_bel_context().items(), columns=cols)

        # skip here if there is no pure symbols with or without BEL context
        if any([df_all.empty, df_bel.empty]):
            return inserted

        edge_types = [
            "controls-transport-of",
            "controls-expression-of",
            "controls-phosphorylation-of",
        ]

        for edge_type in edge_types:
            inserted[edge_type] = 0

            sql = f"""Select id, participant_a, participant_b from
                pathway_commons where interaction_type='{edge_type}'"""
            df_ppi_of = pd.read_sql(sql, self.engine)
            df_join = (
                df_ppi_of.set_index("participant_a")
                .join(df_all.set_index("symbol"))
                .rename(columns={"rid": "rid_a_all"})
                .join(df_bel.set_index("symbol"))
                .reset_index()
                .rename(columns={"rid": "rid_a_bel", "index": "a"})
                .set_index("participant_b")
                .join(df_all.set_index("symbol"))
                .rename(columns={"rid": "rid_b_all"})
                .join(df_bel.set_index("symbol"))
                .reset_index()
                .rename(columns={"rid": "rid_b_bel", "index": "b"})
                .set_index("id")
            )

            df_join["a_is_valid"] = [(x in valid_hgnc_symbols) for x in df_join.a]
            df_join["b_is_valid"] = [(x in valid_hgnc_symbols) for x in df_join.b]

            a_or_b_in_bel = df_join.rid_a_bel.notnull() | df_join.rid_b_bel.notnull()
            both_valid = df_join.a_is_valid & df_join.b_is_valid
            df_both = df_join[(a_or_b_in_bel & both_valid)]

            for pc_id, row in tqdm(
                df_both.iterrows(),
                total=df_both.shape[0],
                desc=f"Update PC {edge_type}",
            ):
                from_rid = self._get_rid(row.rid_a_all, row.a_is_valid, row.a, pure_symbol_rids_dict)
                to_rid = self._get_rid(row.rid_b_all, row.b_is_valid, row.b, pure_symbol_rids_dict)

                if from_rid and to_rid:
                    pathways, pmids, sources = self.get_pathway_pmids_sources(pc_id, pc_pathway_name_rid_dict)
                    value_dict = {
                        "type": edge_type,
                        "sources": sources,
                        "pmids": pmids,
                        "pathways": pathways,
                    }

                    class_name = edge_type.replace("-", "_") + "_pc"
                    self.create_edge(
                        class_name=class_name,
                        from_rid=from_rid,
                        to_rid=to_rid,
                        value_dict=value_dict,
                        ignore_empty_values=True,
                    )
                    inserted[edge_type] += 1

        return inserted

    def get_pathway_pmids_sources(self, pc_id, pc_pathway_name_rid_dict) -> tuple:
        """Return all pathway, PMIDs, and their sources."""
        pc_obj = self.session.query(pc.PathwayCommons).get(pc_id)
        sources = [x.source for x in pc_obj.sources]
        pmids = [x.pmid for x in pc_obj.pmids]
        pathways = [pc_pathway_name_rid_dict[x.name] for x in pc_obj.pathway_names]
        return pathways, pmids, sources



"""Essential API query methods."""

import json
import re
from collections import namedtuple
from enum import Enum
from math import ceil
from typing import Dict, List, Type

from flask import request
from sqlalchemy import inspect, not_
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy.orm.decl_api import DeclarativeMeta
from sqlalchemy.orm.query import Query

from ebel import Bel
from ebel.web.api import RDBMS

Pagination = namedtuple("Pagination", ["page", "page_size", "skip"])


class OrientDbSqlOperator(Enum):
    """String constant definitions."""

    EQUALS = "="
    LIKE = "like"
    IN = "in"
    STARTS_WITH = "like"
    ENDS_WITH = "like"
    CONTAINS = "like"


class DataType(Enum):
    """String constant definitions."""

    STRING = 1
    INTEGER = 2
    NUMBER = 3
    LIST_STRING = 4
    LIST_INTEGER = 5
    LIST_NUMBER = 6


class SubRelObj(Enum):
    """String constant definitions."""

    SUBJECT = "s"
    RELATION = "r"
    OBJECT = "o"


def _get_pagination() -> Pagination:
    """Get page and page_size from request."""
    request_obj = request.args if request.args else json.loads(request.data)
    page_size = request_obj.get("page_size", 10)
    page_size = int(page_size) if (isinstance(page_size, int) or re.search(r"^\d+$", page_size)) else 10
    page = request_obj.get("page", 1)
    page = int(page) if (isinstance(page, int) or re.search(r"^\d+$", page)) else 1
    skip = (page - 1) * page_size
    return Pagination(page=page, page_size=page_size, skip=skip)


def _get_data(
    model: Type[DeclarativeMeta],
    print_sql=False,
    order_by: List[InstrumentedAttribute] = [],
):
    columns: Dict[str, InstrumentedAttribute] = {
        col_name: col_obj for col_name, col_obj in model.__dict__.items() if isinstance(col_obj, InstrumentedAttribute)
    }
    bool_map = {"true": 1, "false": 0}
    request_obj = request.args if request.args else json.loads(request.data)
    params = {k: (bool_map[v] if v in bool_map else v) for k, v in request_obj.items() if k in columns and v}
    like_queries = [columns[k].like(v) for k, v in params.items()]
    query = RDBMS.get_session().query(model).filter(*like_queries)
    return _get_paginated_query_result(query, print_sql=print_sql, order_by=order_by)


def _get_paginated_query_result(
    query: Query,
    return_dict=False,
    print_sql=False,
    order_by: List[InstrumentedAttribute] = [],
):
    """Return paginated query result if sqlalchemy model have as_dict method.

    Method requires page_size and page in `request.args`
    """
    p = _get_pagination()

    if not (p.page and p.page_size):
        return {"error": "Please add page and page_size to your method."}

    number_of_results = query.count()
    limit = int(p.page_size)
    page = int(p.page)
    offset = (page - 1) * limit
    offset = offset if offset <= 100 else 100
    pages = ceil(number_of_results / limit)
    if order_by:
        for col in order_by:
            query = query.order_by(col)
    q = query.limit(limit).offset(offset)

    if print_sql:
        print(q.statement.compile(compile_kwargs={"literal_binds": True}))

    if return_dict:
        results = [row._asdict() for row in q.all()]
    else:
        results = [x.as_dict() for x in q.all()]

    return {
        "page": page,
        "page_size": limit,
        "number_of_results": number_of_results,
        "pages": pages,
        "results": results,
    }


def _get_paginated_ebel_query_result(sql: str, print_sql=False):
    """Return paginated ebel query result. Method requires page_size and page in `request.args`."""
    b = Bel()
    p = _get_pagination()

    if not (p.page and p.page_size):
        return {"error": "Please add page and page_size to your method."}

    number_of_results = b.query_get_dict(f"SELECT count(*) from ({sql})")[0]["count"]
    limit = int(p.page_size)
    page = int(p.page)
    offset = (page - 1) * limit
    offset = offset if offset <= 100 else 100
    pages = ceil(number_of_results / limit)
    sql_paginated = sql + f" skip {offset} limit {limit}"

    if print_sql:
        print(sql_paginated)

    return {
        "page": page,
        "page_size": limit,
        "number_of_results": number_of_results,
        "pages": pages,
        "results": [x for x in b.query_get_dict(sql_paginated)],
    }


def _get_terms_from_model_starts_with(form_field: str, sa_column: InstrumentedAttribute, by="args"):
    return _get_terms_from_model_like(form_field, sa_column, how="starts_with", by=by)


def _get_terms_from_model_ends_with(form_field: str, sa_column: InstrumentedAttribute, by="args"):
    return _get_terms_from_model_like(form_field, sa_column, how="starts_with", by=by)


def _get_terms_from_model_contains(form_field: str, sa_column: InstrumentedAttribute, by="args"):
    return _get_terms_from_model_like(form_field, sa_column, how="contains", by=by)


def _get_terms_from_model_like(form_field: str, sa_column: InstrumentedAttribute, how="like", by="args"):
    """Get terms from SQL Alchemy models field. Alchemy models field should be unique.

    :param form_field: name of form field
    :param sa_column: column of SQL Alchemy model
    :param how: (like, starts_with, ends_with, contains)
    :param by: (args, data)
    """
    search_term, page, page_size = None, None, None
    if by == "args":
        page_size = request.args.get("page_size", 10)
        page = request.args.get("page", 1)
        search_term = request.args.get(form_field, None)
    elif by == "data":
        page_size = request.data.get("page_size", 10)
        page = request.data.get("page", 1)
        search_term = request.data.get(form_field, None)

    if not (search_term and page and page_size):
        return {"error": f"{form_field} is required."}

    if how == "starts_with":
        search_term = f"{search_term}%"
    elif how == "ends_with":
        search_term = f"%{search_term}"
    elif how == "contains":
        search_term = f"%{search_term}%"

    model = sa_column.class_
    primary_key = inspect(model).primary_key[0]
    query = (
        RDBMS.get_session().query(primary_key, sa_column).filter(sa_column.like(f"{search_term}%")).order_by(sa_column)
    )

    number_of_results = query.count()

    limit = int(page_size)
    page = int(page)
    offset = (page - 1) * limit
    offset = offset if offset <= 100 else 100
    pages = ceil(number_of_results / limit)

    return {
        "page": page,
        "page_size": limit,
        "number_of_results": number_of_results,
        "pages": pages,
        "results": {x[1]: x[0] for x in query.limit(limit).offset(offset).all()},
    }


def add_query_filters(query: Query, columns_params: Dict[str, Dict[str, str]], model: DeclarativeMeta):
    """Add optional filters to query."""
    col_filters = []
    for column_name, v in columns_params.items():
        if v.get("how2search") and v.get("value"):
            how2search = v["how2search"]
            print(how2search)
            value = v["value"].strip() if isinstance(v["value"], str) else v["value"]

            column = inspect(model).columns[column_name]
            if how2search in ("exact", "exact_numeric"):
                col_filters.append(column == value)
            elif how2search == "starts_with":
                col_filters.append(column.like(value + "%"))
            elif how2search == "ends_with":
                col_filters.append(column.like("%" + value))
            elif how2search == "contains":
                col_filters.append(column.like("%" + value + "%"))
            elif how2search == "greater_than":
                col_filters.append(column.__gt__(value))
            elif how2search == "greater_equals_than":
                col_filters.append(column.__ge__(value))
            elif how2search == "smaller_than":
                col_filters.append(column.__lt__(value))
            elif how2search == "smaller_equals_than":
                col_filters.append(column.__le__(value))
            elif how2search == "not_equals":
                col_filters.append(column != value)
            elif how2search == "exclude":
                col_filters.append(not_(column.like(value)))
            elif how2search == "between":
                found_2_values = re.search(
                    r"(?P<value_1>[+-]?\d+(\.\d+)?).*?[-,;:/].*?(?P<value_2>[+-]?\d+(\.\d+)?)",
                    value,
                )
                if found_2_values:
                    values = sorted(found_2_values.groupdict().values(), reverse=True)
                    col_filters.append(column.between(*values))

    query = query.filter(*col_filters)
    return query



"""ClinVar."""
import logging
from collections import namedtuple
from typing import Dict, List

import pandas as pd
from pyorientdb import OrientDB
from tqdm import tqdm

from ebel.manager.orientdb import odb_meta, odb_structure, urls
from ebel.manager.orientdb.constants import CLINVAR
from ebel.manager.rdbms.models import clinvar
from ebel.tools import get_disease_trait_keywords_from_config, get_file_path

logger = logging.getLogger(__name__)

Snp = namedtuple(
    "Snp",
    (
        "keyword",
        "phenotype",
        "rs_number",
        "hgnc_id",
        "chromosome",
        "position",
        "clinical_significance",
    ),
)


class ClinVar(odb_meta.Graph):
    """ClinVar (from NCBI)."""

    def __init__(self, client: OrientDB = None):
        """Init ClinVar."""
        self.client = client
        self.biodb_name = CLINVAR
        self.urls = {self.biodb_name: urls.CLINVAR}
        self.file_path = get_file_path(urls.CLINVAR, self.biodb_name)
        super().__init__(
            nodes=odb_structure.clinvar_nodes,
            edges=odb_structure.clinvar_edges,
            indices=odb_structure.clinvar_indices,
            urls=self.urls,
            tables_base=clinvar.Base,
            biodb_name=self.biodb_name,
        )

    def __len__(self):
        return self.number_of_generics

    def __contains__(self, item):
        # TODO: To be implemented
        return True

    def insert_data(self) -> Dict[str, int]:
        """Insert data."""
        inserted = {}
        self.recreate_tables()
        df = pd.read_csv(self.file_path, sep="\t", low_memory=False)
        self._standardize_dataframe(df)
        df.index += 1
        df.index.rename("id", inplace=True)
        df.drop(columns=["phenotype_ids", "phenotype_list", "other_ids"]).to_sql(
            self.biodb_name, self.engine, if_exists="append", chunksize=10000
        )

        df_clinvar__phenotype = (
            df["phenotype_list"]
            .str.split(r"[|;]")
            .explode()
            .to_frame()
            .reset_index()
            .rename(columns={"phenotype_list": "phenotype", "id": "clinvar_id"})
        )
        df_clinvar__phenotype.index += 1
        df_phenotype = pd.DataFrame(df_clinvar__phenotype.phenotype.unique(), columns=["phenotype"])
        df_phenotype.index += 1
        df_phenotype.index.rename("id", inplace=True)
        df_phenotype.to_sql(
            clinvar.ClinvarPhenotype.__tablename__,
            self.engine,
            if_exists="append",
            chunksize=10000,
        )
        inserted.update({clinvar.ClinvarPhenotype.__tablename__: df_phenotype.shape[0]})

        df_clinvar__phenotype = (
            df_clinvar__phenotype.set_index("phenotype")
            .join(df_phenotype.assign(clinvar_phenotype_id=df_phenotype.index).set_index("phenotype"))
            .reset_index()
            .loc[:, ["clinvar_id", "clinvar_phenotype_id"]]
        )

        df_clinvar__phenotype.index += 1
        df_clinvar__phenotype.index.rename("id", inplace=True)
        df_clinvar__phenotype.to_sql(
            "clinvar__phenotype",
            self.engine,
            if_exists="append",
            index=False,
            chunksize=10000,
        )
        inserted.update({clinvar.clinvar__clinvar_phenotype.__dict__["fullname"]: df_clinvar__phenotype.shape[0]})

        clinvar_pheno_medgen = (
            df["phenotype_ids"]
            .str.split(r"[|,;]")
            .explode()
            .str.partition(":")[[0, 2]]
            .rename(columns={0: "db", 2: "identifier"})
        )
        clinvar_pheno_medgen["clinvar_id"] = clinvar_pheno_medgen.index
        clinvar_phenotype_medgen = (
            clinvar_pheno_medgen.set_index("db").loc["MedGen"].reset_index().drop(columns=["db"])
        )
        clinvar_phenotype_medgen.index += 1
        clinvar_phenotype_medgen.index.rename("id", inplace=True)
        clinvar_phenotype_medgen.to_sql(
            clinvar.ClinvarPhenotypeMedgen.__tablename__,
            self.engine,
            if_exists="append",
            chunksize=10000,
        )
        inserted.update({clinvar.ClinvarPhenotypeMedgen.__tablename__: clinvar_phenotype_medgen.shape[0]})

        df_other_ids = (
            df.other_ids.str.split(",")
            .dropna()
            .explode()
            .str.partition(":")[[0, 2]]
            .reset_index()
            .rename(columns={0: "db", 2: "identifier", "id": "clinvar_id"})
            .rename_axis("id")
        )
        df_other_ids.index += 1
        df_other_ids.to_sql(
            clinvar.ClinvarOtherIdentifier.__tablename__,
            self.engine,
            if_exists="append",
            chunksize=10000,
        )
        inserted.update({clinvar.ClinvarOtherIdentifier.__tablename__: df_other_ids.shape[0]})
        return inserted

    def update_bel(self) -> int:
        """Delete and updates all ClinVar edges."""
        logger.info(f"Clear edges for {self.biodb_name.upper()}")
        self.clear_edges()
        logger.info("Clear nodes without edges")
        self.clear_nodes_with_no_edges()

        added_edges = self.update_interactions()
        logger.info(f"Successfully updated interactions for {self.biodb_name.upper()}")
        return added_edges

    def get_disease_snps_dict(self) -> Dict[str, List[Snp]]:
        """Get a dictionary {'disease':[snp,snp,... ]} by disease names."""
        disease_keywords = get_disease_trait_keywords_from_config()

        sql_temp = """Select
            '{keyword}',
            phenotype,
            rs_db_snp as rs_number,
            hgnc_id,
            chromosome,
            start as position,
            clinical_significance
                from clinvar c inner join
                clinvar__phenotype cp on (c.id=cp.clinvar_id) inner JOIN
                clinvar_phenotype p on (cp.clinvar_phenotype_id=p.id)
            where
                p.phenotype like '%%{keyword}%%'
                and rs_db_snp != -1"""

        results = dict()
        for kwd in disease_keywords:
            sql = sql_temp.format(keyword=kwd)
            rows = self.engine.execute(sql)
            results[kwd] = [Snp(*x) for x in rows.fetchall()]

        return results

    def update_interactions(self) -> int:
        """Create SNPs and edges using information from ClinVar."""
        # snp is upstream of a downstream gene
        snp_type = {
            "mapped": "mapped",
            "downstream": "upstream",
            "upstream": "downstream",
        }
        added_edges = 0
        disease_snps_dict = self.get_disease_snps_dict()
        hgnc_id_gene_rid_cache = {}

        for disease, rows in disease_snps_dict.items():
            for snp in tqdm(rows, desc=f"Add has_X_snp_cv edges to BEL for {disease}"):
                if snp.hgnc_id in hgnc_id_gene_rid_cache:
                    gene_mapped_rid = hgnc_id_gene_rid_cache[snp.hgnc_id]
                else:
                    gene_mapped_rid = self._get_set_gene_rid(hgnc_id=snp.hgnc_id)
                    hgnc_id_gene_rid_cache[snp.hgnc_id] = gene_mapped_rid

                if gene_mapped_rid:
                    snp_rid = self.get_create_rid("snp", {"rs_number": "rs" + str(snp.rs_number)})
                    value_dict = {
                        "clinical_significance": snp.clinical_significance,
                        "phenotype": snp.phenotype,
                        "keyword": snp.keyword,
                    }
                    self.create_edge(
                        class_name="has_mapped_snp_cv",
                        from_rid=gene_mapped_rid,
                        to_rid=snp_rid,
                        value_dict=value_dict,
                        if_not_exists=True,
                    )
                    added_edges += 1

                    # fetch all down and upstream gene_rids
                    gene_type_rids = self.get_set_gene_rids_by_position(
                        chromosome=snp.chromosome,
                        position=snp.position,
                        gene_types=["downstream", "upstream"],
                    )

                    for gene_type, gene_rids in gene_type_rids.items():
                        for gene_rid in gene_rids:
                            class_name = f"has_{snp_type[gene_type]}_snp_cv"
                            self.create_edge(
                                class_name=class_name,
                                from_rid=gene_rid,
                                to_rid=snp_rid,
                                value_dict=value_dict,
                            )
                            added_edges += 1
        return added_edges

    def _get_set_gene_rid(self, hgnc_id: str):
        """Insert gene (if not exists) and returns OrientDB @rid."""
        gene_rid = None

        hgnc_rids = self.query_class("hgnc", columns=["symbol"], limit=1, id=hgnc_id)

        if hgnc_rids:
            gene = hgnc_rids[0]
            bel = f'g(HGNC:"{gene["symbol"]}")'
            data = {
                "pure": True,
                "bel": bel,
                "name": gene["symbol"],
                "namespace": "HGNC",
            }
            gene_rid = self.get_create_rid("gene", data, check_for="bel")

        return gene_rid



"""nSIDES API methods."""

from flask import request

from ebel.manager.rdbms.models.nsides import Nsides
from ebel.web.api.ebel.v1 import _get_data, _get_paginated_ebel_query_result


def get_nsides():
    """Get generic NSIDES entry."""
    return _get_data(Nsides)


def get_ebel_relation():
    """Get NSIDES related eBEL relations."""
    conf = {
        "drugbank_id": "out.drugbank_id",
        "drug_label": "out.label",
        "ppr": "prr",
        "mean_reporting_frequency": "mean_reporting_frequency",
        "condition_meddra_id": "in.condition_meddra_id",
        "side_effect": "in.label",
    }
    sql = "SELECT "
    sql += ", ".join([f"{v} as {k}" for k, v in conf.items()])
    sql += " FROM has_side_effect"

    paras = {k: request.args[k] for k in request.args if k in conf}
    if paras:
        wheres = [f'{conf[k].replace(".asString()","")} like "{v}"' for k, v in paras.items()]
        sql += " WHERE " + " AND ".join(wheres)

    return _get_paginated_ebel_query_result(sql)



"""StringDB RDBMS model definition."""

from sqlalchemy import Boolean, Column, Integer, SmallInteger, String
from sqlalchemy.ext.declarative import declarative_base

from ebel.manager.rdbms.models import object_as_dict

Base = declarative_base()


class StringDb(Base):
    """Class definition for the stringdb table."""

    __tablename__ = "stringdb"

    id = Column(Integer, primary_key=True)

    protein1 = Column(String(50), nullable=False)
    protein2 = Column(String(50), nullable=False)
    symbol1 = Column(String(50), nullable=False, index=True)
    symbol2 = Column(String(50), nullable=False, index=True)
    neighborhood = Column(Integer)
    neighborhood_transferred = Column(SmallInteger)
    fusion = Column(SmallInteger)
    cooccurence = Column(SmallInteger)
    homology = Column(SmallInteger)
    coexpression = Column(SmallInteger)
    coexpression_transferred = Column(SmallInteger)
    experiments = Column(SmallInteger, index=True)
    experiments_transferred = Column(SmallInteger)
    database = Column(Integer)
    database_transferred = Column(SmallInteger)
    textmining = Column(SmallInteger)
    textmining_transferred = Column(SmallInteger)
    combined_score = Column(SmallInteger)

    def as_dict(self):
        """Convert object values to dictionary."""
        return object_as_dict(self, exclude=["id"])


class StringDbProtein(Base):
    """Class definition for the stringdb_protein table."""

    __tablename__ = "stringdb_protein"

    id = Column(Integer, primary_key=True)
    string_protein_id = Column(String(50), nullable=False, index=True)
    preferred_name = Column(String(50), nullable=False, index=True)

    def as_dict(self):
        """Convert object values to dictionary."""
        return object_as_dict(self, exclude=["id"])


class StringDbAction(Base):
    """Class definition for the stringdb_action table."""

    __tablename__ = "stringdb_action"
    id = Column(Integer, primary_key=True)
    item_id_a = Column(String(50), nullable=False)
    item_id_b = Column(String(50), nullable=False)
    symbol1 = Column(String(50), nullable=False, index=True)
    symbol2 = Column(String(50), nullable=False, index=True)
    mode = Column(String(20), nullable=False, index=True)
    action = Column(String(20))
    is_directional = Column(Boolean, nullable=False, index=True)
    a_is_acting = Column(Boolean, nullable=False, index=True)
    score = Column(SmallInteger)

    def as_dict(self):
        """Convert object values to dictionary."""
        return object_as_dict(self, exclude=["id"])



"""ClinVar RDBMS model definition."""
from sqlalchemy import Column, ForeignKey, Index, Integer, String, Table, Text
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

from ebel.manager.rdbms.models import object_as_dict

Base = declarative_base()

clinvar__clinvar_phenotype = Table(
    "clinvar__phenotype",
    Base.metadata,
    Column("clinvar_id", Integer, ForeignKey("clinvar.id")),
    Column("clinvar_phenotype_id", Integer, ForeignKey("clinvar_phenotype.id")),
)


class ClinvarPhenotypeMedgen(Base):
    """Class definition for the clinvar_phenotype_medgen table."""

    __tablename__ = "clinvar_phenotype_medgen"
    id = Column(Integer, primary_key=True)

    identifier = Column(String(100), index=True)
    clinvar_id = Column(Integer, ForeignKey("clinvar.id"))
    clinvar = relationship("Clinvar", foreign_keys=[clinvar_id], viewonly=True)


class ClinvarOtherIdentifier(Base):
    """Class definition for the clinvar_other_identifier table."""

    __tablename__ = "clinvar_other_identifier"
    id = Column(Integer, primary_key=True)

    db = Column(String(100), index=True)
    identifier = Column(String(100), index=True)
    clinvar_id = Column(Integer, ForeignKey("clinvar.id"))
    clinvar = relationship("Clinvar", foreign_keys=[clinvar_id], viewonly=True)

    def as_dict(self):
        """Convert object values to dictionary."""
        return {"db": self.db, "identifier": self.identifier}


class Clinvar(Base):
    """Class definition for the clinvar table."""

    __tablename__ = "clinvar"
    id = Column(Integer, primary_key=True)

    allele_id = Column(Integer)
    type = Column(String(100))
    name = Column(String(1000))
    gene_id = Column(Integer, index=True)
    gene_symbol = Column(String(1000))
    hgnc_id = Column(String(100))
    clinical_significance = Column(String(100))
    clin_sig_simple = Column(Integer)
    last_evaluated = Column(String(100))
    rs_db_snp = Column(Integer, index=True)
    nsv_esv_db_var = Column(String(100))
    rcvaccession = Column(String(1000))
    origin = Column(Text)
    origin_simple = Column(Text)
    assembly = Column(String(100), index=True)
    chromosome_accession = Column(Text)
    chromosome = Column(Text)
    start = Column(Integer)
    stop = Column(Integer)
    reference_allele = Column(Text)
    alternate_allele = Column(Text)
    cytogenetic = Column(Text)
    review_status = Column(Text)
    number_submitters = Column(Integer)
    guidelines = Column(Text)
    tested_in_gtr = Column(Text)
    submitter_categories = Column(Integer)
    variation_id = Column(Integer)
    position_vcf = Column(Integer)
    reference_allele_vcf = Column(Text(100000))
    alternate_allele_vcf = Column(Text(100000))

    phenotypeMedgens = relationship("ClinvarPhenotypeMedgen", foreign_keys=[ClinvarPhenotypeMedgen.clinvar_id])
    otherIdentifiers = relationship("ClinvarOtherIdentifier", foreign_keys=[ClinvarOtherIdentifier.clinvar_id])

    phenotypes = relationship("ClinvarPhenotype", secondary=clinvar__clinvar_phenotype)

    __table_args__ = (Index("ix_clinvar__gene_symbol", gene_symbol, mysql_length=500),)

    def as_dict(self):
        """Convert object values to dictionary."""
        clinvar_entry = object_as_dict(self)
        clinvar_entry.update({"phenotypeMedgens": [x.identifier for x in self.phenotypeMedgens]})
        clinvar_entry.update({"otherIdentifiers": [x.as_dict() for x in self.otherIdentifiers]})
        return clinvar_entry


class ClinvarPhenotype(Base):
    """Class definition for the clinvar_phenotype table."""

    __tablename__ = "clinvar_phenotype"
    id = Column(Integer, primary_key=True)
    phenotype = Column(Text)

    clinvars = relationship("Clinvar", secondary=clinvar__clinvar_phenotype, back_populates="phenotypes")

    def as_dict(self):
        """Convert object values to dictionary."""
        return {"phenotype": self.phenotype}



"""HGNC."""
import json
import logging
import re
import sys
from collections import namedtuple
from typing import Dict, Optional

import numpy as np
import pandas as pd
from pyorientdb import OrientDB
from tqdm import tqdm

from ebel.manager.orientdb import odb_meta, odb_structure, urls
from ebel.manager.orientdb.constants import HGNC
from ebel.manager.rdbms.models import hgnc
from ebel.tools import get_file_path

logger = logging.getLogger(__name__)
HgncEntry4Update = namedtuple(
    "HgncEntry4Update",
    ["hgnc_rid", "label", "location", "symbol", "suggested_corrections"],
)


class Hgnc(odb_meta.Graph):
    """HGNC class definition."""

    def __init__(self, client: Optional[OrientDB] = None):
        """Init HGNC."""
        self.client = client
        self.biodb_name = HGNC
        self.urls = {self.biodb_name: urls.HGNC_JSON, "human_ortholog": urls.HCOP_GZIP}
        super().__init__(
            generics=odb_structure.hgnc_generics,
            tables_base=hgnc.Base,
            indices=odb_structure.hgnc_indices,
            nodes=odb_structure.hgnc_nodes,
            urls=self.urls,
            biodb_name=self.biodb_name,
        )

    def __contains__(self, hgnc_id: object) -> bool:
        """Test existence of hgnc_id."""
        if isinstance(hgnc_id, int):
            hgnc_id = "HGNC:{}".format(hgnc_id)
        r = self.execute("Select count(*) from bel where hgnc.id = '{}' limit 1".format(hgnc_id))
        return bool(len(r[0].oRecordData["count"]))

    def __len__(self):
        """Count number of hgnc links in BEL graph."""
        r = self.execute("Select count(*) from bel where hgnc IS NOT NULL")
        return r[0].oRecordData["count"]

    def __repr__(self) -> str:
        """Represent HGNC."""
        template = "{{BioDatabase:Hgnc}}[url:{url}, nodes:{nodes}, generics:{generics}]"
        representation = template.format(url=self.urls, nodes=self.number_of_nodes, generics=self.number_of_generics)
        return representation

    def insert_data(self) -> Dict[str, int]:
        """Check if files missing for download or generic table empty. If True then insert data."""
        inserted = dict()
        inserted["hgnc"] = self.import_hgnc()
        inserted["hgnc_rdbms"] = self.import_hgnc_into_rdbms()
        inserted["human_orthologs"] = self.insert_orthologs()
        self.session.commit()
        return inserted

    def import_hgnc_into_rdbms(self) -> int:
        """Insert HGNC database into RDBMS."""
        logger.info("Insert HGNC database into RDBMS.")
        file_path = get_file_path(self.urls[self.biodb_name], self.biodb_name)

        with open(file_path, "r", encoding="utf8") as hgnc_file:
            raw_content = hgnc_file.read()
            string_encode = raw_content.encode("ascii", "ignore")  # Convert unicode chars to ascii
            hgnc_content = json.loads(string_encode.decode())

        df = pd.DataFrame(hgnc_content["response"]["docs"])

        self._standardize_dataframe(df)
        columns = [
            "hgnc_id",
            "version",
            "bioparadigms_slc",
            "cd",
            "cosmic",
            "date_approved_reserved",
            "date_modified",
            "date_name_changed",
            "date_symbol_changed",
            "ensembl_gene_id",
            "entrez_id",
            "homeodb",
            "horde_id",
            "imgt",
            "iuphar",
            "lncipedia",
            "lncrnadb",
            "location",
            "location_sortable",
            "locus_group",
            "locus_type",
            "merops",
            "mirbase",
            "name",
            "orphanet",
            "snornabase",
            "status",
            "symbol",
            "ucsc_id",
            "uuid",
            "vega_id",
            "agr",
        ]

        df["id"] = pd.to_numeric(df.hgnc_id.str.split(":").str[1])
        df.set_index("id", inplace=True)
        df[columns].to_sql(hgnc.Hgnc.__tablename__, self.engine, if_exists="append")

        df.hgnc_id = pd.to_numeric(df.hgnc_id.str.split(":").str[1])

        for df_col, model, m_col in (
            ("prev_symbol", hgnc.PrevSymbol, None),
            ("alias_symbol", hgnc.AliasSymbol, None),
            ("alias_name", hgnc.AliasName, None),
            ("ccds_id", hgnc.Ccds, "identifier"),
            ("ena", hgnc.Ena, "identifier"),
            ("enzyme_id", hgnc.Enzyme, "ec_number"),
            ("gene_group", hgnc.GeneGroupName, "name"),
            ("gene_group_id", hgnc.GeneGroupId, "identifier"),
            ("uniprot_ids", hgnc.UniProt, "accession"),
            ("rna_central_id", hgnc.RnaCentral, "identifier"),
            ("rgd_id", hgnc.Rgd, "identifier"),
            ("refseq_accession", hgnc.RefSeq, "accession"),
            ("pubmed_id", hgnc.PubMed, "pmid"),
            ("prev_name", hgnc.PrevName, None),
            ("omim_id", hgnc.Omim, "identifier"),
            ("mgd_id", hgnc.Mgd, "identifier"),
            ("lsdb", hgnc.Lsdb, "identifier"),
        ):
            df_1n_table = df[[df_col, "hgnc_id"]].explode(df_col).dropna()
            if m_col:
                df_1n_table.rename(columns={df_col: m_col}, inplace=True)
            df_1n_table.to_sql(model.__tablename__, self.engine, if_exists="append", index=False)

        return df.shape[0]

    def import_hgnc(self) -> int:
        """Import HGNC into OrientDB."""
        # if new hgnc is imported all hgnc links should be reset and hgnc table should be empty
        self.execute("Update genetic_flow set hgnc=null")
        self.execute("Delete from hgnc")

        file_path = get_file_path(self.urls[self.biodb_name], self.biodb_name)

        with open(file_path, "r", encoding="utf8") as hgnc_file:
            raw_content = hgnc_file.read()
            string_encode = raw_content.encode("ascii", "ignore")  # Convert unicode chars to ascii
            hgnc_content = json.loads(string_encode.decode())

        rows = hgnc_content["response"]["docs"]
        df = pd.DataFrame(rows)
        df = self._standardize_dataframe(df)
        df.rename(columns={"hgnc_id": "id"}, inplace=True)
        df = df.where(pd.notnull(df), None)
        sql_temp = "INSERT INTO `hgnc` content {}"

        new_entries = df.to_dict("records")

        for row in tqdm(new_entries, desc=f"Import {self.biodb_name.upper()}"):
            sql = sql_temp.format({k: v for k, v in row.items() if v})

            try:
                self.execute(sql)

            except Exception as e:
                print(e)
                print(sql)
                sys.exit()

        return len(rows)

    def insert_orthologs(self) -> int:
        """Import orthologs from European Molecular Biology Laboratory."""
        # table_name = hgnc.HumanOrtholog.__tablename__
        table_name = "human_ortholog"

        file_path = get_file_path(self.urls["human_ortholog"], self.biodb_name)
        used_columns = [
            "hgnc_id",
            "human_entrez_gene",
            "human_ensembl_gene",
            "human_symbol",
            "ortholog_species",
            "ortholog_species_entrez_gene",
            "ortholog_species_ensembl_gene",
            "ortholog_species_db_id",
            "ortholog_species_symbol",
            "support",
        ]
        df = pd.read_csv(file_path, sep="\t", low_memory=False, usecols=used_columns)
        df.index += 1
        df.index.rename("id", inplace=True)
        df.replace("-", np.nan, inplace=True)
        df.to_sql(table_name, self.engine, chunksize=100000, if_exists="append")
        return df.shape[0]

    def get_basic_entry_by_symbol(self, symbol: str) -> Optional[HgncEntry4Update]:
        """Return HgncEntry4Update object."""
        sql_name = """Select
                        @rid.asString() as hgnc_rid,
                        status,
                        name as label,
                        location,
                        symbol
                    from
                        hgnc where symbol='{}' limit 1""".format(
            symbol
        )
        result = self.execute(sql_name)

        if result:
            data = result[0].oRecordData
            status = data.pop("status")
            if status == "Entry Withdrawn":
                data["suggested_corrections"] = f"{status}: Please use correct HGNC symbol"
            else:
                data["suggested_corrections"] = None

            if "location" in data:
                data["location"] = self.get_location(data["location"])
            else:
                data["location"] = None
            return HgncEntry4Update(**data)

    @staticmethod
    def get_location(location: str) -> dict:
        """Return formatted version of gene location."""
        regex = r"(?P<chromosome>(\d+|X|Y|mitochondria))(((?P<arm>p|q)?(?P<region>\d+))?(\.(?P<band>\d+)))?"
        locus_found = re.match(regex, location)
        # overwrite data['location'] with structured info in dict
        if locus_found:
            locus_dict = locus_found.groupdict()
            lr = locus_dict["region"]
            band = locus_dict["band"]
            locus_dict["region"] = int(lr) if lr else lr
            locus_dict["band"] = int(band) if band else band
            location_dict = locus_dict
        else:
            location_dict = {"unknown_schema": location}
        return location_dict

    def get_bel_symbols_without_hgnc_link(self):
        """Return set of all gene symbols in database without a link to HGNC."""
        sql_symbols = "Select distinct(name) as symbol from bio_object where namespace='HGNC' and hgnc IS NULL"
        return {x.oRecordData["symbol"] for x in self.execute(sql_symbols)}

    def get_bel_symbols_all(self):
        """Return set of all gene symbols in database."""
        sql_symbols = "Select distinct(name) as symbol from bio_object where namespace='HGNC'"
        return {x.oRecordData["symbol"] for x in self.execute(sql_symbols)}

    def get_correct_symbol(self, symbol: str):
        """Checks if symbol is valid otherwise checks previsous symbols."""
        result_in_symbol = self.session.query(hgnc.Hgnc).filter(hgnc.Hgnc.symbol == symbol).first()
        if not result_in_symbol:
            result_in_prev_symbol = (
                self.session.query(hgnc.PrevSymbol).filter(hgnc.PrevSymbol.prev_symbol == symbol).first()
            )
            if result_in_prev_symbol:
                symbol = result_in_prev_symbol.hgnc.symbol
            else:
                symbol = None
        return symbol

    def correct_wrong_symbol(self, symbol, bel_symbols_all: set):
        """Corrects the symbol of the node and relinks all edges to existing node if needed."""
        result = self.session.query(hgnc.PrevSymbol).filter_by(prev_symbol=symbol).first()
        if result:
            correct_symbol = result.hgnc.symbol
            if correct_symbol not in bel_symbols_all:
                sql = f"Select @rid.asString(), bel from bio_object where namespace='HGNC' and name='{symbol}'"
                for row in self.execute(sql):
                    r = row.oRecordData
                    rid = r["rid"]
                    correct_bel = re.sub(rf'(?<=:")({symbol})(?=")', correct_symbol, r["bel"])
                    sql = f"UPDATE {rid} SET bel='{correct_bel}', name='{correct_symbol}'"
                    self.execute(sql)
            else:
                # TODO: correction only possible if all edges are relinked to the correct node and old node is deleted
                pass

    def update_bel(self) -> int:
        """Update links in protein, rna and gene nodes to HGNC."""
        # check if hgnc is in ODB
        if not self.execute("Select count(*) as num from hgnc")[0].oRecordData["num"]:
            self.import_hgnc()

        bel_symbols_all = self.get_bel_symbols_all()
        symbols_without_hgnc = self.get_bel_symbols_without_hgnc_link()
        hgnc_symbols = {x[0] for x in self.session.query(hgnc.Hgnc.symbol).all()}

        for wrong_symbol in symbols_without_hgnc - hgnc_symbols:
            self.correct_wrong_symbol(wrong_symbol, bel_symbols_all)

        new_bel_symbols = self.get_bel_symbols_without_hgnc_link()

        updated = 0
        for symbol in tqdm(new_bel_symbols, desc="Update HGNC"):
            if self.update_nodes_by_symbol(symbol):
                updated += 1
        return updated

    def update_gene(
        self,
        hgnc_rid: str,
        label: str,
        location: str,
        hgnc_symbol: str,
        suggested_corrections: str,
    ) -> int:
        """Update genes in OrientDB and returns number of updates."""
        suggest = (
            ", suggested_corrections={{'wrong name': {}}}".format(suggested_corrections)
            if suggested_corrections
            else ""
        )
        sql_temp = """Update gene set
                        hgnc = {hgnc_rid},
                        label= {label},
                        location={location} {suggest}
                    where namespace = 'HGNC' and name = '{hgnc_symbol}'"""
        sql = sql_temp.format(
            hgnc_rid=hgnc_rid,
            label=json.dumps(label),
            location=json.dumps(location),
            hgnc_symbol=hgnc_symbol,
            suggest=suggest,
        )
        return self.execute(sql)[0]

    def update_rna(self, hgnc_rid: str, label: str, hgnc_symbol: str, suggested_corrections: str) -> int:
        """Update RNAs in OrientDB and returns number of updates."""
        suggest = (
            ", suggested_corrections={{'wrong name': {}}}".format(suggested_corrections)
            if suggested_corrections
            else ""
        )
        sql_temp = """Update rna set hgnc = {hgnc_rid},label={label} {suggest}
                      where namespace = 'HGNC' and name = '{hgnc_symbol}'"""
        sql = sql_temp.format(
            hgnc_rid=hgnc_rid,
            label=json.dumps(label),
            hgnc_symbol=hgnc_symbol,
            suggest=suggest,
        )
        return self.execute(sql)[0]

    def update_protein(self, hgnc_rid: str, label: str, hgnc_symbol: str, suggested_corrections: str) -> int:
        """Update proteins in OrientDB and returns number of updates."""
        suggest = (
            ", suggested_corrections={{'wrong name': {}}}".format(suggested_corrections)
            if suggested_corrections
            else ""
        )
        sql_temp = """Update protein set hgnc = {hgnc_rid},label={label} {suggest}
                      where namespace = 'HGNC' and name = '{hgnc_symbol}'"""
        sql = sql_temp.format(
            hgnc_rid=hgnc_rid,
            label=json.dumps(label),
            hgnc_symbol=hgnc_symbol,
            suggest=suggest,
        )
        return self.execute(sql)[0]

    def update_nodes_by_symbol(self, symbol) -> dict:
        """Update all nodes by HGNC symbol."""
        hgnc = self.get_basic_entry_by_symbol(symbol)

        if hgnc:
            suggest = json.dumps(hgnc.suggested_corrections) if hgnc.suggested_corrections else None

            num_update_genes = self.update_gene(
                hgnc_symbol=hgnc.symbol,
                hgnc_rid=hgnc.hgnc_rid,
                label=hgnc.label,
                location=hgnc.location,
                suggested_corrections=suggest,
            )
            num_update_rnas = self.update_rna(
                hgnc_symbol=hgnc.symbol,
                hgnc_rid=hgnc.hgnc_rid,
                label=hgnc.label,
                suggested_corrections=suggest,
            )
            num_update_proteins = self.update_protein(
                hgnc_symbol=hgnc.symbol,
                hgnc_rid=hgnc.hgnc_rid,
                label=hgnc.label,
                suggested_corrections=suggest,
            )
            return {
                "genes": num_update_genes,
                "rnas": num_update_rnas,
                "proteins": num_update_proteins,
            }

    def get_symbol_entrez_dict(self) -> Dict[str, int]:
        """Return dictionary with gene symbols as keys and entrez IDs as values."""
        query = self.session.query(hgnc.Hgnc.symbol, hgnc.Hgnc.entrez_id).filter(hgnc.Hgnc.entrez_id.isnot(None))
        return {r.symbol: r.entrez_id for r in query.all()}

    def update_interactions(self) -> int:
        """Abstract method."""
        pass



"""This module allows to parse BEL scripts."""

import codecs
import json
import logging
import re
from collections import OrderedDict, defaultdict
from copy import copy
from typing import Any, Iterable, List

from lark import Lark
from lark.exceptions import UnexpectedInput, UnexpectedToken
from pandas import DataFrame

from ebel.constants import GRAMMAR_BEL_PATH, GRAMMAR_START_LINE
from ebel.errors import BelSyntaxError
from ebel.manager.models import load_grammar
from ebel.transformers import _BelTransformer

# TODO: check all strings if they can be stored in constants

# Following trees (rules in grammar) have one 1 value. This list is use in creation of json file
trees_with_one_value = (
    "document_name",
    "document_version",
    "document_description",
    "document_copyright",
    "document_authors",
    "document_licences",
    "document_contact_info",
    "keyword",
    "anno_keyword",
    "c_type",
    "c_title",
    "c_ref",
    "evidence",
    "frag_range",
    "frag_descriptor",
    "hgvs",
    "gene_fusion_range",
    "rna_fusion_range",
    "protein_fusion_range",
    "document_keywords",
)


# Exclude this token types if json of created
exclude_token_types = ("OB", "CB", "QM", "COLON", "COMMA", "OCB", "CCB")

logger = logging.getLogger(__name__)


def camel_case(name: str) -> str:
    """Camel case a string."""
    first, *rest = name.split("_")
    return first + "".join(word.capitalize() for word in rest)


def tupleit(lst: list) -> tuple:
    """Cast all lists in nested list to tuple."""
    return tuple(map(tupleit, lst)) if isinstance(lst, list) else lst


def to_tuple(py_obj: Any) -> tuple:
    """Create from nested python object a nested tuple where all obj are casted to tuples."""
    lst = []

    # check if instance is dict, str, Iterable or something different
    if isinstance(py_obj, dict):
        for k, v in py_obj.items():
            lst.append((k, to_tuple(v)))

    elif not isinstance(py_obj, str) and isinstance(py_obj, Iterable):
        lst.append(tuple(to_tuple(x) for x in py_obj))

    else:
        return py_obj

    return tuple(lst)


def get_values(childs: list, num_expected_values=0) -> list:
    """Get values from token list and exclude exclude_token_types.

    Parameters
    ----------
    childs : list
        token list.
    num_expected_values : type
        Number of expeted values.

    Returns
    -------
    list
        List of token values.

    """
    values = [x.value for x in childs if x.type not in exclude_token_types]
    if num_expected_values:
        values = values + [""] * (num_expected_values - len(values))
    return values


def append_to_list(py_tree, rule_name, value):
    """Append value to py_tree.

    Append value to py_tree list entry if list entry is dict with rule_name;
    otherwise create a new dict entry with rule_name as key and value as avlue.

    Parameters
    ----------
    py_tree : list
        python object.
    rule_name : str
        rule name from grammar.
    value : any
        any type from Lark.Tree.

    """
    rule_name_index = -1
    rule_name_exists = False
    for i, x in enumerate(py_tree):
        if isinstance(x, dict) and list(x.keys()) == [rule_name]:
            rule_name_index = i
            rule_name_exists = True

    if rule_name_exists:
        if value not in py_tree[rule_name_index][rule_name]:
            py_tree[rule_name_index][rule_name] = sorted(py_tree[rule_name_index][rule_name] + [value])
    else:
        py_tree.append({rule_name: [value]})


def bel_to_json(tree) -> str:
    """Return BEL script as JSON string.

    Parameters
    ----------
    tree: OrderedDict
        Parsed BEL relations represented as a tree, output from Lark.

    Returns
    -------
    str
        BEL tree as a JSON string.
    """
    return json.dumps(tree, indent=1)


class Pattern:
    """Pattern in BEL scripts."""

    def __init__(self):
        """Initialize Pattern with dicttionaries for patterns and errors.

        Returns
        -------
        type
            Description of returned object.
        """
        self.__pattern = {}
        self.errors = defaultdict(list)

    def set_pattern(self, namespace, pattern):
        """Set namespace as regular expression pattern.

        Parameters
        ----------
        namespace : str
            `namespace` use in BEL script.
        pattern : str
            `pattern` as regular expression used in BEL statements.

        Returns
        -------
        type
            Description of returned object.
        """
        self.__pattern[namespace] = pattern

    def check_entry(self, namespace, entry, line, column):
        """Short summary.

        Parameters
        ----------
        namespace : str
            Description of parameter `namespace`.
        entry : str
            Description of parameter `entry`.
        line : int
            Description of parameter `line`.
        column : int
            Description of parameter `column`.

        Returns
        -------
        type
            Description of returned object.
        """
        # TODO: check if this follows error conventions > see Error classes
        if not re.search(self.__pattern[namespace], entry):
            self.errors[namespace].append((entry, line, column))

    @property
    def namespaces(self):
        """Return all namespace for patterns.

        Returns
        -------
        list
            return a list of namespace linked to patterns.
        """
        return self.__pattern.keys()


class _BELParser:
    """class manage the parsing with python lib lark-parser."""

    @staticmethod
    def check_bel_script_line_by_line(bel_script_path: str, bel_version: str) -> List[str]:
        """Check BEL script line by line.

        Parameters
        ----------
        bel_script_path : str
            Path to BEL file.
        bel_version : str
            BEL grammar version.

        Returns
        -------
        List[str]
            List of errors found in BEL script.
        """
        logger.info("Start syntax check for {} line by line with grammar BEL {}".format(bel_script_path, bel_version))
        errors = []

        grammar = load_grammar(GRAMMAR_BEL_PATH[str(bel_version)])

        parser = Lark(grammar, start=GRAMMAR_START_LINE, parser="lalr", lexer="contextual")

        cached_line = ""
        with codecs.open(bel_script_path, "r", encoding="utf-8") as fd:
            lines = fd.readlines()
            if not re.search("(\n|\r|\r\n)$", lines[-1]):
                lines[-1] += "\n"

            num_and_lines = OrderedDict(enumerate(lines, 1))

            for line_num, line in copy(num_and_lines).items():
                if re.search(r"\\\s*(\n|\r|\r\n)$", line):
                    num_and_lines[line_num + 1] = (
                        num_and_lines[line_num].strip()[:-1] + " " + num_and_lines[line_num + 1]
                    )
                    del num_and_lines[line_num]

            for line_number, line in num_and_lines.items():
                try:
                    if line.endswith("\\n"):
                        cached_line = line
                        continue

                    elif cached_line:
                        if not line.endswith("\\n"):
                            try:
                                cached_line += line
                                parser.parse(cached_line)

                            except (UnexpectedInput, UnexpectedToken) as exc1:
                                errors.append(BelSyntaxError(exc1, line_number, cached_line.strip()))

                            cached_line = ""
                            continue

                        else:
                            # TODO: check if last line will be parsed?
                            cached_line += line
                            continue

                    parser.parse(line)

                except (UnexpectedInput, UnexpectedToken) as exc2:
                    errors.append(BelSyntaxError(exc2, line_number, line))

        return errors

    def check_bel_script(self, bel_script_path: str, bel_version: str, force_new_db: bool = False) -> dict:
        """Check file with BEL script for syntax correctness.

        Parameters
        ----------
        bel_script_path : str
            Path to BEL script.
        bel_version : str
            BEL version 1 or 2.
        force_new_db : bool
            Force to create a new database if already exists.

        Returns
        -------
        dict
            Dictionary with results.
        """
        grammar = load_grammar(GRAMMAR_BEL_PATH[bel_version])

        if bel_version.startswith("2"):  # TODO change this hardcoded value
            transformer = _BelTransformer()

        else:
            logger.error(f"Transformer for version {bel_version} not implemented", exc_info=True)
            raise

        parser = Lark(
            grammar,
            start="script",
            parser="lalr",
            lexer="contextual",
            transformer=transformer,
        )

        with codecs.open(bel_script_path, "r", encoding="utf-8") as fd:
            bel_content = fd.read() + "\n"

        tree = None
        warnings = None

        try:
            tree = parser.parse(bel_content)
            errors = transformer.cache.errors
            warnings = transformer.cache.warnings

        except UnexpectedInput as exc:
            logger.info("UnexpectedInput exception in check_bel_script: %s" % exc)
            errors = self.check_bel_script_line_by_line(bel_script_path, bel_version)

        return {"errors": errors, "tree": tree, "warnings": warnings}


def write_error_report(data_frame: DataFrame, file_path: str) -> None:
    """Short summary.

    Parameters
    ----------
    data_frame : type
        Description of parameter `data_frame`.
    file_path : str
        Description of parameter `file_path`.

    Returns
    -------
    type
        Description of returned object.
    """
    if file_path.endswith(".xlsx"):
        data_frame.to_excel(file_path)
    else:
        data_frame.to_csv(file_path)


def write_warning_report(data_frame: DataFrame, file_path: str) -> None:
    """Short summary.

    Parameters
    ----------
    data_frame : DataFrame
        DataFrame with validation results.
    file_path : str
        file path to error report.

    """
    if file_path.endswith(".xlsx"):
        data_frame.to_excel(file_path)
    else:
        data_frame.to_csv(file_path)


def check_bel_script_line_by_line(bel_script_path, error_report_file_path, bel_version):
    """Check statements in file or string for correct.

    result['trees'][line_number] = {'statement': statement, 'tree': tree}
    result['errors'][line_number] = {'statement': statement, 'error': ex}

    # can be used as comments
    empty lines will be skipped
    every statement should be in a new line

    :param str bel_script_path: path to BEL script
    :param str error_report_file_path: path to report
    :param str bel_version: BEL version
    :return: dict
    """
    parser = _BELParser()
    data_frame = parser.check_bel_script_line_by_line(bel_script_path, bel_version)
    if error_report_file_path:
        write_error_report(data_frame=data_frame, file_path=error_report_file_path)
    else:
        return data_frame


def check_bel_script(bel_script_path, bel_version, force_new_db=False) -> dict:
    """Check BEL script.

    Parameters
    ----------
    bel_script_path : str
        Description of parameter `bel_script_path`.
    bel_version : str
        BEL syntax version.
    force_new_db : bool
        if true, force to create new database

    Returns
    -------
    type
        Description of returned object.
    """
    bel_parser = _BELParser()
    result = bel_parser.check_bel_script(
        bel_script_path=bel_script_path,
        force_new_db=force_new_db,
        bel_version=bel_version,
    )

    return result



"""DisGeNet RDBMS model definition."""
from sqlalchemy import BigInteger, Column, Float, ForeignKey, Integer, String
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

from ebel.manager.rdbms.models import object_as_dict

Base = declarative_base()


class DisgenetGene(Base):
    """Class definition for the disgenet_gene table."""

    __tablename__ = "disgenet_gene"
    id = Column(Integer, primary_key=True)

    gene_id = Column(Integer, ForeignKey("disgenet_gene_symbol.gene_id"))
    gene_symbol = relationship("DisgenetGeneSymbol", back_populates="gene_disease_pmid_associations")
    disease_id = Column(String(100), ForeignKey("disgenet_disease.disease_id"))
    disease = relationship("DisgenetDisease", foreign_keys=[disease_id])
    score = Column(Float)
    pmid = Column(BigInteger)
    source_id = Column(Integer, ForeignKey("disgenet_source.id"))
    source = relationship("DisgenetSource", foreign_keys=[source_id])

    def as_dict(self):
        """Convert object values to dictionary."""
        rs = object_as_dict(self, exclude=["id", "source_id"])
        rs.update(
            {
                "gene_symbol": self.gene_symbol.gene_symbol,
                "disease_name": self.disease.disease_name,
                "source": self.source.source,
            }
        )
        return rs


class DisgenetGeneSymbol(Base):
    """Class definition for the disgenet_gene_symbol table."""

    __tablename__ = "disgenet_gene_symbol"
    gene_id = Column(Integer, primary_key=True)
    gene_symbol = Column(String(50), index=True)

    gene_disease_pmid_associations = relationship("DisgenetGene", back_populates="gene_symbol")

    def as_dict(self):
        """Convert object values to dictionary."""
        return object_as_dict(self)


class DisgenetVariant(Base):
    """Class definition for the disgenet_variant table."""

    __tablename__ = "disgenet_variant"
    id = Column(Integer, primary_key=True)

    snp_id = Column(String(20), index=True)
    chromosome = Column(String(2))
    position = Column(BigInteger)
    disease_id = Column(String(100), ForeignKey("disgenet_disease.disease_id"))
    disease = relationship("DisgenetDisease", foreign_keys=[disease_id])
    score = Column(Float)
    pmid = Column(BigInteger, index=True)
    source_id = Column(Integer, ForeignKey("disgenet_source.id"))
    source = relationship("DisgenetSource", foreign_keys=[source_id])

    def as_dict(self):
        """Convert object values to dictionary."""
        rs = object_as_dict(self, exclude=["id", "source_id"])
        rs.update({"disease_name": self.disease.disease_name, "source": self.source.source})
        return rs


class DisgenetDisease(Base):
    """Class definition for the disgenet_disease table."""

    __tablename__ = "disgenet_disease"
    disease_id = Column(String(100), primary_key=True)
    disease_name = Column(String(255), index=True)

    def as_dict(self):
        """Convert object values to dictionary."""
        return object_as_dict(self)


class DisgenetSource(Base):
    """Class definition for the disgenet_source table."""

    __tablename__ = "disgenet_source"
    id = Column(Integer, primary_key=True)
    source = Column(String(100), index=True)

    def as_dict(self):
        """Convert object values to dictionary."""
        return object_as_dict(self, exclude=["id"])



"""Warning class definitions."""
from collections import OrderedDict

TEMPLATE = (
    "{error_class}\tkeyword:{keyword}\tentry:{entry}\tline:{line_number}\tcolumn:{column}" "\turl:{url}\thint:{hint}"
)


class _Warning:
    """Base class for all errors."""

    def __init__(self):
        self.class_name = self.__class__.__name__
        self.value_dict = OrderedDict(
            [
                ("warning_class", self.class_name),
                ("url", None),
                ("keyword", None),
                ("entry", None),
                ("line_number", None),
                ("column", None),
                ("hint", None),
            ]
        )

    def to_dict(self):
        """Format the properties of error into a dictionary."""
        raise NotImplementedError("to_dict have to be implemented in {}".format(self.__class__.__name__))

    def to_string(self) -> str:
        self.value_dict.update(self.to_dict())
        return TEMPLATE.format(**self.value_dict)

    def __str__(self):
        return self.to_string()


class AlsoUsedInOtherNamespace(_Warning):
    """Error in entry links to a namespace defined in the header but not exists in namespace url."""

    def __init__(self, keyword: str, entry: str, line_number: int, column: int, hint: str):
        """Initialize warning class.

        :param keyword: Error class type.
        :param hint: Sentence or string to give the user in the error report to help solve error.
        :param line_number: Line number where error was found.
        :param column: Column number where error was found.
        :param entry: String representation where error occurred.
        """
        super(AlsoUsedInOtherNamespace, self).__init__()

        self.ns_keyword = keyword
        self.entry = entry
        self.line_number = line_number
        self.column = column
        self.hint = hint

    def to_dict(self) -> dict:
        """Format the properties of error into a dictionary."""
        return {
            "error_class": self.class_name,
            "keyword": self.ns_keyword,
            "entry": self.entry,
            "line_number": self.line_number,
            "column": self.column,
            "hint": self.hint,
        }



"""Root init for eBEL."""
from . import cache, constants, errors, parser, transformers
from .manager.orientdb.biodbs.bel import Bel

__version__ = "1.0.33"

__title__ = "e(BE:L)"
__description__ = "Validation and extension of biomedical knowledge graphs"
__url__ = "https://github.com/e-bel/ebel"

__author__ = "Christian Ebeling"
__email__ = "christian.ebeling@scai.fraunhofer.de"

__license__ = "?"
__copyright__ = """Copyright (c) 2021 Christian Ebeling, Fraunhofer Institute for Algorithms and Scientific
Computing SCAI, Schloss Birlinghoven, 53754 Sankt Augustin, Germany"""

project_name = __title__



"""GWAS Catalog API unit tests."""

from .conftest import format_response_data
from .constants import NUM_RESULTS, PAGE_SIZE, RESULTS


class TestGwasCatalog:
    def test_get_gwas_catalog(self, client):
        response = client.get(
            "api/v1/gwas_catalog/gwas_catalog?date_added_to_catalog=2013-10-24&pubmedid=23563609&first_author=Wheeler%20E&date=2013-04-07&journal=Nat%20Genet&link=www.ncbi.nlm.nih.gov%2Fpubmed%2F23563609&study=Genome-wide%20SNP%20and%20CNV%20analysis%20identifies%20common%20and%20low-frequency%20variants%20associated%20with%20severe%20early-onset%20obesity.&disease_trait=Obesity%20%28early%20onset%20extreme%29&initial_sample_size=1%2C509%20European%20ancestry%20cases%2C%205%2C380%20European%20ancestry%20controls&replication_sample_size=971%20European%20ancestry%20cases%2C%201%2C990%20European%20ancestry%20controls&region=16q12.2&chr_id=16&chr_pos=53767042&reported_gene_s=FTO&mapped_gene=FTO&strongest_snp_risk_allele=rs1421085-C&snp=rs1421085&snp_id_current=1421085&context=intron_variant&risk_allele_frequency=0.41&p_value=3e-28&pvalue_mlog=27.5229&or_or_beta=1.44&_95_ci_text=%5B1.35-1.54%5D&platform_snps_passing_qc=Affymetrix%20%5B~%202000000%5D%20%28imputed%29&cnv=N",
            content_type="application/json",
        )
        output = format_response_data(response)
        expected_results = {
            "_95_ci_text": "[1.35-1.54]",
            "chr_id": "16",
            "chr_pos": "53767042",
            "cnv": "N",
            "context": "intron_variant",
            "date": "2013-04-07",
            "date_added_to_catalog": "2013-10-24",
            "disease_trait": "Obesity (early onset extreme)",
            "downstream_gene_distance": None,
            "downstream_gene_id": None,
            "first_author": "Wheeler E",
            "id": 146661,
            "initial_sample_size": "1,509 European ancestry cases, 5,380 European ancestry controls",
            "intergenic": 0,
            "journal": "Nat Genet",
            "link": "www.ncbi.nlm.nih.gov/pubmed/23563609",
            "mapped_gene": "FTO",
            "merged": 0,
            "or_or_beta": 1.44,
            "p_value": 3e-28,
            "p_value_text": None,
            "platform_snps_passing_qc": "Affymetrix [~ 2000000] (imputed)",
            "pubmedid": 23563609,
            "pvalue_mlog": 27.5229,
            "region": "16q12.2",
            "replication_sample_size": "971 European ancestry cases, 1,990 European ancestry controls",
            "reported_gene_s": "FTO",
            "risk_allele_frequency": "0.41",
            "snp": "rs1421085",
            "snp_genes": ["ENSG00000140718"],
            "snp_id_current": "1421085",
            "strongest_snp_risk_allele": "rs1421085-C",
            "study": "Genome-wide SNP and CNV analysis identifies common and low-frequency variants associated with severe early-onset obesity.",
            "upstream_gene_distance": None,
            "upstream_gene_id": None,
        }
        assert output[NUM_RESULTS] == 1
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results



"""DrugBank RDBMS model definition."""

from sqlalchemy import Column, Date, ForeignKey, Integer, String, Text
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

Base = declarative_base()


class Drugbank(Base):
    """Class definition for the drugbank table."""

    __tablename__ = "drugbank"
    id = Column(Integer, primary_key=True)
    drugbank_id = Column(String(10), index=True)
    name = Column(String(255))
    description = Column(Text)
    cas_number = Column(String(20))
    unii = Column(String(20))
    state = Column(String(20))
    indication = Column(Text)
    pharmacodynamics = Column(Text)
    toxicity = Column(Text)
    metabolism = Column(Text)
    absorption = Column(Text)
    half_life = Column(Text)
    route_of_elimination = Column(Text)
    volume_of_distribution = Column(Text)
    clearance = Column(Text)
    mechanism_of_action = Column(Text)
    fda_label = Column(Text)

    references = relationship("Reference", back_populates="drugbank", cascade="save-update")
    synonyms = relationship("Synonym", back_populates="drugbank", cascade="save-update")
    targets = relationship("Target", back_populates="drugbank", cascade="save-update")
    external_identifiers = relationship("ExternalIdentifier", back_populates="drugbank", cascade="save-update")
    product_names = relationship("ProductName", back_populates="drugbank", cascade="save-update")
    drug_interactions = relationship("DrugInteraction", back_populates="drugbank", cascade="save-update")
    statuses = relationship("Status", back_populates="drugbank", cascade="save-update")
    patents = relationship("Patent", back_populates="drugbank", cascade="save-update")
    pathways = relationship("Pathway", back_populates="drugbank", cascade="save-update")

    def __str__(self):
        """Class string definition."""
        return self.drugbank_id

    # TODO: add drug_interaction
    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            "drugbank_id": self.drugbank_id,
            "name": self.name,
            "description": self.description,
            "cas_number": self.cas_number,
            "toxicity": self.toxicity,
            "indication": self.indication,
            "pharmacodynamics": self.pharmacodynamics,
            "metabolism": self.metabolism,
            "absorption": self.absorption,
            "half_life": self.half_life,
            "route_of_elimination": self.route_of_elimination,
            "volume_of_distribution": self.volume_of_distribution,
            "clearance": self.clearance,
            "mechanism_of_action": self.mechanism_of_action,
            "fda_label": self.fda_label,
            "references": [x.pmid for x in self.references],
            "pathways": [x.smpdb_id for x in self.pathways],
            "patents": [x.as_dict() for x in self.patents],
            "targets": [x.as_dict() for x in self.targets],
            "product_names": [x.name for x in self.product_names],
            "external_identifiers": [x.as_dict() for x in self.external_identifiers],
            "statuses": [x.status for x in self.statuses],
        }


class Pathway(Base):
    """Class definition for the drugbank_pathway table."""

    __tablename__ = "drugbank_pathway"
    id = Column(Integer, primary_key=True)
    smpdb_id = Column(String(255))

    drugbank_id = Column(Integer, ForeignKey("drugbank.id"))
    drugbank = relationship("Drugbank", back_populates="pathways")

    def __str__(self):
        return self.smpdb_id

    def as_dict(self):
        """Convert object values to dictionary."""
        return {"smpdb_id": self.smpdb_id, "drugbank_id": self.drugbank.drugbank_id}


class Patent(Base):
    """Class definition for the drugbank_patent table."""

    __tablename__ = "drugbank_patent"
    id = Column(Integer, primary_key=True)
    number = Column(String(255))
    country = Column(String(255))
    approved = Column(Date)
    expires = Column(Date)
    pediatric_extension = Column(String(255))

    drugbank_id = Column(Integer, ForeignKey("drugbank.id"))
    drugbank = relationship("Drugbank", back_populates="patents")

    def __str__(self):
        return self.number

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            "number": self.number,
            "country": self.country,
            "approved": self.approved.strftime("%Y-%m-%d"),
            "expires": self.expires.strftime("%Y-%m-%d"),
            "pediatric_extension": self.pediatric_extension,
            "drugbank_id": self.drugbank.drugbank_id,
        }


class Status(Base):
    """Class definition for the drugbank_status table."""

    __tablename__ = "drugbank_status"
    id = Column(Integer, primary_key=True)
    status = Column(String(20), index=True)

    drugbank_id = Column(Integer, ForeignKey("drugbank.id"))
    drugbank = relationship("Drugbank", back_populates="statuses")

    def __str__(self):
        return self.status

    def as_dict(self):
        """Convert object values to dictionary."""
        return {"smpdb_id": self.status, "drugbank_id": self.drugbank.drugbank_id}


class ExternalIdentifier(Base):
    """Class definition for the drugbank_external_identifier table."""

    __tablename__ = "drugbank_external_identifier"
    id = Column(Integer, primary_key=True)
    resource = Column(String(255), index=True)
    identifier = Column(String(255), index=True)

    drugbank_id = Column(Integer, ForeignKey("drugbank.id"))
    drugbank = relationship("Drugbank", back_populates="external_identifiers")

    def __str__(self):
        return self.identifier

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            "resource": self.resource,
            "identifier": self.identifier,
            "drugbank_id": self.drugbank.drugbank_id,
        }


class Reference(Base):
    """Class definition for the drugbank_reference table."""

    __tablename__ = "drugbank_reference"
    id = Column(Integer, primary_key=True)
    pmid = Column(Integer)

    drugbank_id = Column(Integer, ForeignKey("drugbank.id"))
    drugbank = relationship("Drugbank", back_populates="references")

    def __str__(self):
        return self.pmid

    def as_dict(self):
        """Convert object values to dictionary."""
        return {"pmid": self.pmid, "drugbank_id": self.drugbank.drugbank_id}


class Target(Base):
    """Class definition for the drugbank_target table."""

    __tablename__ = "drugbank_target"
    id = Column(Integer, primary_key=True)
    uniprot = Column(String(20), index=True)
    action = Column(String(50), index=True)
    known_action = Column(String(20), index=True)

    drugbank_id = Column(Integer, ForeignKey("drugbank.id"))
    drugbank = relationship("Drugbank", back_populates="targets")

    def __str__(self):
        return self.uniprot

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            "uniprot": self.uniprot,
            "action": self.action,
            "known_action": self.known_action,
            "drugbank_id": self.drugbank.drugbank_id,
        }


class DrugInteraction(Base):
    """Class definition for the drugbank_drug_interaction table."""

    __tablename__ = "drugbank_drug_interaction"
    id = Column(Integer, primary_key=True)
    drugbank_id = Column(String(10), index=True)
    name = Column(Text)
    description = Column(Text)

    db_id = Column(Integer, ForeignKey("drugbank.id"))  # exception because drugbank_id is already a field
    drugbank = relationship("Drugbank", back_populates="drug_interactions")

    def __str__(self):
        return self.drugbank_id

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            "drugbank_id": self.drugbank_id,
            "name": self.name,
            "description": self.description,
            "interactor_drugbank_id": self.drugbank.drugbank_id,
        }


class ProductName(Base):
    """Class definition for the drugbank_product_name table."""

    __tablename__ = "drugbank_product_name"
    id = Column(Integer, primary_key=True)
    name = Column(Text)

    drugbank_id = Column(Integer, ForeignKey("drugbank.id"))
    drugbank = relationship("Drugbank", back_populates="product_names")

    def __str__(self):
        return self.name

    def as_dict(self):
        """Convert object values to dictionary."""
        return {"drugbank_id": self.drugbank.drugbank_id, "name": self.name}


class Synonym(Base):
    """Class definition for the drugbank_synonym table."""

    __tablename__ = "drugbank_synonym"
    id = Column(Integer, primary_key=True)
    synonym = Column(Text)

    drugbank_id = Column(Integer, ForeignKey("drugbank.id"))
    drugbank = relationship("Drugbank", back_populates="synonyms")

    def __str__(self):
        return self.synonym

    def as_dict(self):
        """Convert object values to dictionary."""
        return {"drugbank_id": self.drugbank.drugbank_id, "synonym": self.synonym}



"""Generic BEL relation API methods."""
import cgi
import io
import json
import re
from collections import Counter, defaultdict, namedtuple
from copy import deepcopy
from enum import Enum
from math import ceil
from pathlib import Path
from typing import Any, Dict, List, NamedTuple, Optional, Set, Tuple, Union

import pandas as pd
import requests
import xmltodict
from flask import make_response, request, send_file, send_from_directory
from graphviz import Digraph

from ebel import Bel
from ebel.manager.orientdb.odb_structure import (get_columns,
                                                 get_node_view_labels)
from ebel.validate import validate_bel_file
from ebel.web.api.ebel.v1 import DataType, OrientDbSqlOperator, _get_pagination

PathLengthDict = Dict[int, List[Dict[str, list]]]
PathLength = int
EdgeInfo = dict
NodeInfo = dict
Rid = str
RidList = List[str]
EdgePathsByLength = Dict[PathLength, RidList]
ErrorMessage = Dict[str, str]


class EnumExtension(Enum):
    """Extension enumeration class."""

    @classmethod
    def has_value(cls, value) -> bool:
        """Check if value present."""
        return any([x.value == value for x in cls])

    @classmethod
    def has_name(cls, name) -> bool:
        """Check if has name."""
        return any([x.name == name for x in cls])


BELishEdge = namedtuple("Edge", ["name", "direction", "params_str"])

edge_colours = {
    "increases": "limegreen",
    "directly_increases": "springgreen4",
    "decreases": "orangered",
    "directly_decreases": "red3",
    "rate_limiting_step_of": "lightslateblue",
    "regulates": "lightblue3",
    "causes_no_change": "yellow2",
    "positive_correlation": "darkolivegreen4",
    "negative_correlation": "coral3",
}

node_colours = {
    "protein": "lightblue1",
    "complex": "khaki1",
    "component": "aquamarine",
    "rna": "goldenrod1",
    "gene": "lightslateblue",
    "activity": "palegreen3",
    "abundance": "darksalmon",
    "pathology": "palegreen",
    "drug_db": "yellow1",
    "biological_process": "snow",
}


def validate_uploaded_bel_file(file, errorOutput: str = "download", forceJson: bool = False):
    """Validate the uploaded BEL file and return either the excel or JSON file.

    errorOutput : str
        Can either be "download" (downloads error report as excel), "json" (results as table in JSON), or "html" to
        return error report table directly as html
    forceJson : bool
        If True, compiles JSON file even if there are errors
    """
    filename = file.filename
    error_report_file = f"{filename}.errors.xlsx"
    json_file = f"{filename}.json"

    with open(filename, "wb") as tmpf:
        tmpf.write(file.read())

    validate_bel_file(filename, reports=error_report_file, force_json=forceJson)

    if Path(error_report_file).is_file() and not forceJson:
        if errorOutput == "download":
            # Read data into memory so we can delete file after
            return_data = io.BytesIO()
            with open(error_report_file, "rb") as errorf:
                return_data.write(errorf.read())

            return_data.seek(0)  # (after writing, cursor will be at last byte, so move it to start)

            response = send_file(
                return_data,
                as_attachment=True,
                mimetype="application/vnd.ms-excel",
                attachment_filename=error_report_file,
            )

        elif errorOutput == "json":
            error_df = pd.read_excel(error_report_file, index_col=[0])
            # content = error_df.to_json(orient='records')
            content = error_df.to_dict(orient="records")
            response = {"type": "error-report", "format": "json", "content": content}

        elif errorOutput == "html":
            error_df = pd.read_excel(error_report_file, index_col=[0])
            content = error_df.to_html()
            response = {"type": "error-report", "format": "html", "content": content}

        else:
            raise ValueError(f"errorOutput must be either 'download', 'json', or 'html', not {errorOutput}")

        Path(error_report_file).unlink()

    else:  # Successful JSON
        with open(json_file, "r") as jf:
            content = json.load(jf)
            response = {"type": "success-report", "format": "json", "content": content}
        Path(json_file).unlink()

    Path(filename).unlink()

    return response


class Column:
    """Column definition class."""

    def __init__(
        self,
        form_name: str,
        column: str,
        sql_operator: OrientDbSqlOperator = OrientDbSqlOperator.EQUALS,
        data_type: DataType = DataType.STRING,
        value: str = None,
        show_in_results: bool = True,
        switch_where_terms=False,
    ):
        """Init method for column.

        Parameters
        ----------
        form_name
        column
        sql_operator
        data_type
        """
        self.column = column
        self.display_column = column + ".asString()" if "@" in column else column
        self.sql_operator = sql_operator
        self.form_name = form_name
        self.data_type = data_type
        self.value = value
        self.show_in_results = show_in_results
        self.switch_where_terms = switch_where_terms

    def set_search_term(self, value: str):
        """Return value for a given search term."""
        if value:
            self.value = value.strip()

    def get_sql(self):
        """Build the SQL query."""
        column = f"{self.column}.asString()" if "@" in self.column else self.column
        return f"{column} as {self.form_name}"

    def __str__(self):
        return f"""<Column: form_name={self.form_name};
        column={self.column}; sql_operator={self.sql_operator}; data_type={self.data_type}>"""


bel_relation_default_columns: List[Column] = [
    Column("subject_rid", "out.@rid"),
    Column("subject_node_class", "out.@class"),
    Column("subject_namespace", "out.namespace"),
    Column("subject_name", "out.name", OrientDbSqlOperator.LIKE),
    Column("subject_bel", "out.bel", OrientDbSqlOperator.LIKE),
    Column(
        "subject_gene_symbol_involved_in",
        "out.involved_genes",
        OrientDbSqlOperator.IN,
        DataType.LIST_STRING,
    ),
    Column(
        "subject_other_involved_in",
        "out.involved_other",
        OrientDbSqlOperator.IN,
        DataType.LIST_STRING,
    ),
    Column("relation_rid", "@rid"),
    Column("relation", "@class"),
    Column("evidence", "evidence", OrientDbSqlOperator.LIKE),
    Column(
        "citation_full_journal_name",
        "citation.full_journal_name",
        OrientDbSqlOperator.LIKE,
    ),
    Column("citation_pub_date", "citation.pub_date"),
    Column("citation_pub_year", "citation.pub_year"),
    Column("citation_last_author", "citation.last_author", OrientDbSqlOperator.LIKE),
    Column("citation_type", "citation.type"),
    Column(
        "author_in_author_list",
        "citation.author_list",
        OrientDbSqlOperator.IN,
        DataType.LIST_STRING,
    ),
    Column("title", "citation.title", OrientDbSqlOperator.LIKE),
    Column("doi", "citation.doi"),
    Column("object_rid", "in.@rid"),
    Column("object_node_class", "in.@class"),
    Column("object_namespace", "in.namespace"),
    Column("object_name", "in.name", OrientDbSqlOperator.LIKE),
    Column("object_bel", "in.bel", OrientDbSqlOperator.LIKE),
    Column(
        "object_gene_symbol_involved_in",
        "in.involved_genes",
        OrientDbSqlOperator.IN,
        DataType.LIST_STRING,
    ),
    Column(
        "object_other_involved_in",
        "in.involved_other",
        OrientDbSqlOperator.IN,
        DataType.LIST_STRING,
    ),
]

Pagination = namedtuple("Pagination", ["page", "page_size", "skip"])


class Query:
    """Generic class for creating a SQL query."""

    def __init__(self, odb_class: str, columns: List[Column]):
        """Init method for hte Query class."""
        self.odb_class: str = odb_class
        self.columns: List[Column] = columns
        self.ebel = Bel()
        self.where = self.get_where()

    @staticmethod
    def get_pagination() -> Pagination:
        """Separate results into pages of a specific length."""
        page_size = request.args.get("page_size", "10")
        page_size = int(page_size) if re.search(r"^\d+$", page_size) else 10
        page_size = 10 if page_size >= 100 else page_size
        page = request.args.get("page", "1")
        page = int(page) if re.search(r"^\d+$", page) else 1
        skip = (page - 1) * page_size
        return Pagination(page=page, page_size=page_size, skip=skip)

    def get_where(self):
        """Generic filter execution method."""
        where = ""
        wheres = []
        for col in self.columns:
            if col.value:
                if col.column.endswith("@rid"):
                    if "," in col.value:
                        rids = [x.strip() for x in col.value.split(",") if re.search(r"^#\d+:\d+$", x.strip())]
                        rids_str = "[" + ",".join(rids) + "]"
                        wheres.append(f"{col.column} in {rids_str}")
                    else:
                        rid = col.value.strip()
                        if rid:
                            wheres.append(f"{col.column} = {rid}")
                elif col.column != "@class":
                    if col.data_type in [DataType.STRING, DataType.LIST_STRING]:
                        value = f'"{col.value}"'
                    else:
                        value = col.value

                    if col.data_type in [
                        DataType.LIST_STRING,
                        DataType.LIST_NUMBER,
                        DataType.LIST_INTEGER,
                    ]:
                        wheres.append(f"{value} {col.sql_operator.value} {col.column}")
                    else:
                        if col.switch_where_terms:
                            wheres.append(f"{value} {col.sql_operator.value} {col.column}")
                        else:
                            wheres.append(f"{col.column} {col.sql_operator.value} {value}")
        if wheres:
            where = " WHERE " + " AND ".join(wheres)
        return where

    @property
    def sql(self):
        """Generic sql execution method."""
        select = "SELECT "
        select += ", ".join([f"{sw.display_column} as {sw.form_name}" for sw in self.columns if sw.show_in_results])
        select += " FROM " + self.odb_class
        sql = select + self.where
        return sql

    def get_number_of_results(self):
        """Count number of results."""
        sql = "SELECT count(*) FROM " + self.odb_class + self.where
        return self.ebel.query_get_dict(sql)[0]["count"]

    def get_result(self, pagination: Optional[Pagination] = None):
        """Return total number of results."""
        if pagination:
            p = pagination
        else:
            p = self.get_pagination()
            if not (p.page and p.page_size):
                return {"error": "Please add page and page_size to your method."}

        number_of_results = self.get_number_of_results()
        pages = ceil(number_of_results / p.page_size)
        sql_paginated = self.sql + f" skip {p.skip} limit {p.page_size}"
        # print(sql_paginated)

        return {
            "page": p.page,
            "page_size": p.page_size,
            "number_of_results": number_of_results,
            "pages": pages,
            "results": [x for x in self.ebel.query_get_dict(sql_paginated)],
        }


def _get_where_by_how(column: str, value: str, how_to_search: str):
    how_to_search = how_to_search if SearchType.has_value(how_to_search) else SearchType.EXACT.value
    value_by_how = {
        SearchType.EXACT.value: f" = '{value}'",
        SearchType.CONTAINS.value: f" like '%{value}%'",
        SearchType.CASE_SENSITIVE.value: f" like '{value}'",
        SearchType.STARTS_WITH.value: f" like '{value}%'",
        SearchType.ENDS_WITH.value: f" like '%{value}'",
        SearchType.GREATER_THAN.value: f" > {value}",
        SearchType.GREATER_OR_EQUALS_THAN.value: f" >= {value}",
        SearchType.SMALLER_THAN.value: f" < {value}",
        SearchType.SMALLER_OR_EQUALS_THAN.value: f" <= {value}",
    }
    return column + value_by_how[how_to_search]


def get_node_class_bel_name_ns():
    """Return node class, BEL name, and namespace."""
    default_args = (
        "bel",
        "node_name",
        "namespace",
        "node_class",
        "how_bel",
        "how_name",
    )
    args = {x: "" for x in default_args}
    filtered_request_args = {k: v for k, v in request.args.items() if k in default_args}
    args.update(filtered_request_args)
    namespaces = _get_node_namespace_list(**args)
    node_classes = _get_node_class_list(**args)
    suggested_node_names = _get_suggested_node_names(**args)
    suggested_bels = _get_suggested_bels(**args)
    return {
        "namespaces": namespaces,
        "node_classes": node_classes,
        "suggested_node_names": suggested_node_names,
        "suggested_bels": suggested_bels,
    }


def _get_suggested_bels(
    bel: str,
    node_name: str,
    node_class: str,
    namespace: str,
    how_name: str,
    how_bel: str,
):
    node_class = node_class if node_class else "bel"
    sql = f"Select bel from {node_class}"
    where = []

    if bel:
        where.append(_get_where_by_how(column="bel", value=bel, how_to_search=how_bel))

    if namespace:
        where.append(f"namespace = '{namespace}'")

    if node_name:
        where.append(_get_where_by_how("name", node_name, how_name))

    if where:
        sql += " where " + " and ".join(where)

    sql += " order by bel limit 30"
    print(sql)
    return [y for y in [x.oRecordData.get("bel") for x in Bel().execute(sql)] if y is not None]


def _get_suggested_node_names(
    bel: str,
    node_name: str,
    node_class: str,
    namespace: str,
    how_name: str,
    how_bel: str,
):
    node_class = node_class if node_class else "bel"

    sql = f"Select name from {node_class} where "
    where = []
    where.append(_get_where_by_how(column="name", value=node_name, how_to_search=how_name))

    if namespace:
        where.append(f"namespace = '{namespace}'")

    if bel:
        where.append(_get_where_by_how("bel", bel, how_bel))

    sql += " and ".join(where) + " group by name order by name limit 30"

    return [x.oRecordData["name"] for x in Bel().execute(sql)]


def _get_node_namespace_list(
    bel: str,
    node_name: str,
    namespace: str,
    node_class: str,
    how_name: str,
    how_bel: str,
):
    """Get first names from BEL nodes (by namespace and node_class)."""
    if not namespace:
        node_class = node_class if node_class else "bel"
        sql = f"Select namespace from {node_class} where namespace is not null "
        if node_name:
            sql += " and " + _get_where_by_how(column="name", value=node_name, how_to_search=how_name)
        if bel:
            sql += " and " + _get_where_by_how(column="bel", value=bel, how_to_search=how_bel)
        sql += " group by namespace order by namespace"
        # print(sql)
        return [x.oRecordData["namespace"] for x in Bel().execute(sql)]
    else:
        return [namespace]


def _get_node_class_list(bel: str, node_name: str, node_class: str, namespace: str, how_name: str, how_bel):
    if not node_class:
        sql = "Select @class.asString() as node_class from bel"
        where = []
        if node_name or namespace or bel:
            if node_name:
                where.append(_get_where_by_how(column="name", value=node_name, how_to_search=how_name))
            if namespace:
                where.append(f"namespace = '{namespace}'")
            if bel:
                where.append(_get_where_by_how(column="bel", value=bel, how_to_search=how_bel))
            sql += " where " + " and ".join(where)
        sql += " group by @class order by @class"
        # print(sql)
        return [x.oRecordData["node_class"] for x in Bel().execute(sql)]
    else:
        return [node_class]


def get_namespaces():
    """Get ordered list of namespaces."""
    sql = "Select distinct(namespace) as namespace from bel where namespace is not null order by namespace"
    print(sql)
    return [x.oRecordData["namespace"] for x in Bel().execute(sql)]


def get_node_classes():
    """Get ordered list of node classes."""
    sql = "Select distinct(@class) as node_class from bel order by node_class"
    print(sql)
    return [x.oRecordData["node_class"] for x in Bel().execute(sql)]


def get_bel_relations_by_pmid():
    """Return BEL relations by PMID."""
    columns: List[Column] = [
        Column("subject_rid", "out.@rid"),
        Column("subject_node_class", "out.@class"),
        Column("subject_namespace", "out.namespace"),
        Column("subject_name", "out.name", OrientDbSqlOperator.LIKE),
        Column("subject_bel", "out.bel", OrientDbSqlOperator.LIKE),
        Column(
            "subject_gene_symbol_involved_in",
            "out.involved_genes",
            OrientDbSqlOperator.IN,
            DataType.LIST_STRING,
        ),
        Column(
            "subject_other_involved_in",
            "out.involved_other",
            OrientDbSqlOperator.IN,
            DataType.LIST_STRING,
        ),
        Column("relation_rid", "@rid"),
        Column("relation", "@class"),
        Column("evidence", "evidence", OrientDbSqlOperator.LIKE),
        Column("object_rid", "in.@rid"),
        Column("object_node_class", "in.@class"),
        Column("object_namespace", "in.namespace"),
        Column("object_name", "in.name", OrientDbSqlOperator.LIKE),
        Column("object_bel", "in.bel", OrientDbSqlOperator.LIKE),
        Column(
            "object_gene_symbol_involved_in",
            "in.involved_genes",
            OrientDbSqlOperator.IN,
            DataType.LIST_STRING,
        ),
        Column(
            "object_other_involved_in",
            "in.involved_other",
            OrientDbSqlOperator.IN,
            DataType.LIST_STRING,
        ),
    ]
    column_pmid = Column("pmid", "pmid", data_type=DataType.INTEGER, value=request.args.get("pmid"))
    columns.append(column_pmid)
    sql_builder = Query("bel_relation", columns)
    return sql_builder.get_result(Pagination(1, 1000, 0))


def get_edge_by_annotation() -> list:
    """Return list of edges with a given annotation."""
    columns = deepcopy(bel_relation_default_columns)
    annotation_key = request.args.get("annotation_key")
    annotation_term = request.args.get("annotation_term")
    if annotation_key and annotation_term:
        column = Column(
            "annotation_key",
            f"annotation['{annotation_key}']",
            sql_operator=OrientDbSqlOperator.IN,
            value=annotation_term,
            switch_where_terms=True,
        )
        columns.append(column)

    sql_builder = Query("bel_relation", columns)
    return sql_builder.get_result()


def get_edge_rids():
    """Get edge OrientDB rids."""
    subject_rid = request.args.get("subject_rid")
    relation_rid = request.args.get("relation_rid")
    object_rid = request.args.get("object_rid")
    document_rid = request.args.get("document_rid")
    columns = [
        Column(form_name="subject_rid", column="in.@rid", value=subject_rid),
        Column(form_name="relation_rid", column="@rid", value=relation_rid),
        Column(form_name="object_rid", column="out.@rid", value=object_rid),
        Column(form_name="document_rid", column="document.@rid", value=document_rid),
    ]
    sql_builder = Query("bel_relation", columns)
    return sql_builder.get_result()


def get_annotation_keys():
    """Return annotation keys."""
    sql = """Select value as annotation_key, count(*) as number_of_edges from
          (Select expand(annotation.keys()) as mesh from bel_relation
          where annotation.mesh is not null) group by value order by number_of_edges desc"""
    return [x.oRecordData for x in Bel().execute(sql)]


def get_mesh_terms_statistics_by_node_rid():
    """Return MeSH term stats by node rID."""
    rid = request.args.get("node_rid")
    direction = request.args.get("direction")
    limit = request.args.get("limit")
    sql = f"""Select list(annotation.mesh) as mesh_terms FROM
    (traverse {direction}E() FROM {rid} MAXDEPTH 1) where @rid!={rid} and annotation.mesh is not null"""
    res = Bel().query_get_dict(sql)
    if "mesh_terms" in res[0]:
        res_dict = Counter(res[0]["mesh_terms"])
        mesh_counter_list = [
            {"mesh_term": x[0], "count": x[1]}
            for x in sorted(res_dict.items(), key=lambda item: item[1], reverse=True)
        ]
        return mesh_counter_list[: int(limit)] if limit else mesh_counter_list
    return []


def get_annotation_terms():
    """Get the annotation terms."""
    annotation_key = request.args.get("annotation_key")
    if annotation_key:
        sql = (
            "Select value as annotation_term, count(*) as number_of_edges from "
            f"(Select expand(annotation['{annotation_key}']) as mesh from bel_relation "
            "where annotation.mesh is not null) group by value order by number_of_edges desc"
        )
        return [x.oRecordData for x in Bel().execute(sql)]


def get_edges():
    """Return data for a BEL relation edge."""
    columns = deepcopy(bel_relation_default_columns)

    for column in columns:
        column.set_search_term(request.args.get(column.form_name))

    relation = request.args.get("relation", "bel_relation")
    sql_builder = Query(relation, columns)

    return sql_builder.get_result()


def get_nodes() -> dict:
    """Return list of nodes with a given namespace."""
    b = Bel()
    where_list: List[str] = []
    params = {k: v for k, v in request.args.items() if k in ["namespace", "name"] and v}

    if request.args.get("pure") == "true":
        params.update(pure=True)

    conn2bel_rel = request.args.get("connected_to_bel_relation")
    if conn2bel_rel:
        conn2bel_rel_dir = request.args.get("connected_to_bel_relation_direction", "both")
        where_list.append(f"{conn2bel_rel_dir}('{conn2bel_rel}').size()>0")

    conn2ebel_rel = request.args.get("connected_to_ebel_relation")
    if conn2ebel_rel:
        conn2ebel_rel_dir = request.args.get("connected_to_ebel_relation_direction", "both")
        where_list.append(f"{conn2ebel_rel_dir}('{conn2ebel_rel}').size()>0")

    node_class = request.args.get("node_class")
    p = _get_pagination()
    number_of_results = b.query_class(class_name=node_class, columns=["count(*)"], with_rid=False, **params)[0][
        "count"
    ]
    pages = ceil(number_of_results / p.page_size)
    results = b.query_class(
        class_name=node_class,
        columns=[
            "namespace",
            "name",
            "bel",
            "pure",
            "involved_genes",
            "involved_other",
        ],
        skip=p.skip,
        limit=p.page_size,
        where_list=tuple(where_list),
        print_sql=True,
        **params,
    )
    return {
        "page": p.page,
        "page_size": p.page_size,
        "number_of_results": number_of_results,
        "pages": pages,
        "results": results,
    }


def _get_rid() -> Optional[str]:
    """Get rID."""
    rid = request.args.get("rid")
    if rid:
        rid = rid.strip()
        if re.search(r"#\d+:\d+", rid):
            return rid


def get_edge_statistics_by_rid():
    """Return edge statistics for given rID."""
    rid = request.args.get("rid")
    direction = request.args.get("direction", "both")  # in, out or both
    sql = "Select @class, count(*) from (traverse {dir}E() FROM {rid} MAXDEPTH 1) where @rid!={rid} group by @class"
    res = Bel().query_get_dict(sql.format(dir=direction, rid=rid))
    return res


def get_by_rid() -> Optional[str]:
    """Return BEL node by rid."""
    result_dict = {}
    rid = _get_rid()
    if rid:
        b = Bel()
        for key, value in b.client.record_load(rid).oRecordData.items():
            if isinstance(value, (str, int, float, dict)):
                result_dict[key] = value
            elif isinstance(value, list):
                if all([isinstance(x, (str, int, float)) for x in value]):
                    result_dict[key] = value
        return result_dict


def get_adjacent_nodes_by_rid() -> list:
    """Return neighboring nodes of given rID."""
    # d := direction
    # od := oposite direction
    rid = _get_rid()
    relation = request.args.get("relation", "bel_relation")
    sql_temp = (
        "Select '{d}' as direction, @rid.asString() as edge_rid, @class.asString() "
        "as edge_class, {d}.@rid.asString() as node_rid, {d}.@class.asString()as node_class , "
        "{d}.bel as bel, {d}.name as name, {d}.namespace as namespace, {d}.involved_genes as "
        f"involved_genes, {{d}}.involved_other as involved_other from {relation} "
        f"where {{od}}.@rid = {rid}"
    )

    direction = request.args.get("direction", "both")
    if rid:
        sql_in = sql_temp.format(d="in", od="out")
        sql_out = sql_temp.format(d="out", od="in")
        if direction == "in":
            sql = sql_in
        elif direction == "out":
            sql = sql_out
        else:
            sql = f"select expand($c) let $a = ({sql_in}), $b = ({sql_out}), $c = unionAll( $a, $b )"
        return [x.oRecordData for x in Bel().execute(sql)]


def get_number_of_edges() -> int:
    """Return the number of edges."""
    b = Bel()
    relation = request.args.get("relation", "E")
    r = b.execute(f"Select count(*) as number_of_edges from {relation} limit 1")
    return r[0].oRecordData["number_of_edges"]


def get_citation_by_pmid() -> dict:
    """Return the number of edges."""
    b = Bel()
    pmid = request.args.get("pmid")
    r = b.execute(f"Select citation from bel_relation where pmid = {pmid} limit 1")
    return r[0].oRecordData["citation"]


def get_abstract_by_pmid():
    """Return abstract by PMID."""
    pmid = request.args.get("pmid")
    url = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=pubmed&id={pmid}&retmode=XML&rettype=abstract"
    r = requests.get(url.format(pmid=pmid))
    d = xmltodict.parse(r.text)
    return d["PubmedArticleSet"]["PubmedArticle"]["MedlineCitation"]["Article"]["Abstract"]["AbstractText"]


def get_number_of_nodes() -> int:
    """Return the number of edges."""
    b = Bel()
    node_class = request.args.get("node_class", "V")
    pure = request.args.get("pure")
    where_pure = "where pure = true" if pure else ""
    sql = f"Select count(*) as number_of_nodes from {node_class} {where_pure} limit 1"
    r = b.execute(sql)
    return r[0].oRecordData["number_of_nodes"]


def get_pure_rid() -> Optional[str]:
    """Return None or the rID from the node class."""
    b = Bel()
    node_class = request.args.get("node_class", "protein")
    namespace = request.args.get("namespace", "HGNC")
    name = request.args.get("name")

    if name:
        sql = (
            f"Select @rid.asString() as rid from {node_class} "
            f"where name='{name}' and pure=true and namespace='{namespace}' limit 1"
        )
        return b.execute(sql)[0].oRecordData["rid"]


class Position(EnumExtension):
    """Why is there a class for defining constants."""

    FIRST = "first"
    LAST = "last"
    INSIDE = "inside"


class SearchType(EnumExtension):
    """SearchType constants."""

    EXACT = "exact"
    CONTAINS = "contains"
    CASE_SENSITIVE = "case_insensitive"
    STARTS_WITH = "starts_with"
    ENDS_WITH = "ends_with"
    GREATER_THAN = "greater_than"
    GREATER_OR_EQUALS_THAN = "greater_or_equals_than"
    SMALLER_THAN = "smaller_than"
    SMALLER_OR_EQUALS_THAN = "smaller_or_equals_than"


class MatchEdge:
    """Class to construct the edge portions of a MATCH query."""

    def __init__(
        self,
        edge_class: str,
        multiple_edge_classes: str,
        mesh_terms: List[str],
        pmids: List[int],
    ):
        """Init method."""
        self.position: Optional[Position] = None
        self.node_class = "bel"
        self.edge_class = edge_class
        self.pmids = pmids
        self.mesh_terms = mesh_terms
        self.multiple_edge_classes = multiple_edge_classes

    def set_last(self, node_class: Optional[str]):
        """Set object to the LAST position."""
        self.position = Position.LAST
        if node_class:
            self.node_class = node_class

    def get_edge(self, alias_number: int) -> str:
        """Return edge based on alias number."""
        mesh_or = ""
        if self.mesh_terms:
            mesh_or = " OR ".join(["'" + x.replace("'", "") + "' in annotation.mesh" for x in self.mesh_terms])
            mesh_or = f"({mesh_or})"

        pmids_in = ""
        if self.pmids:
            if len(self.pmids) == 1:
                pmids_in = f"pmid = {self.pmids[0]}"
            else:
                pmids_in = "pmid in " + str(self.pmids)
            pmids_in = f"{pmids_in}"

        where = " AND ".join([x for x in [mesh_or, pmids_in] if x])

        where_str = f"where:({where}), " if any([mesh_or, pmids_in]) else ""
        e_class_multi, e_class_single = self.get_edge_classes()

        return f".outE({e_class_multi}){{{e_class_single}{where_str}as:e{alias_number}}}.inV()"

    def get_edge_classes(self):
        """Return edge classes of match query."""
        """"""
        e_class_single = ""
        e_class_multi = ""

        if isinstance(self.multiple_edge_classes, str) and self.multiple_edge_classes.strip():
            edge_classes = [x.strip() for x in self.multiple_edge_classes.split(",") if x.strip()]
            if len(edge_classes) == 1:
                e_class_single = f"class:{edge_classes[0]}, "
            elif len(edge_classes) > 1:
                edge_classes = ["'" + re.sub(r"\W+", "", x) + "'" for x in edge_classes]
                e_class_multi = ",".join(edge_classes)

        elif isinstance(self.edge_class, str) and self.edge_class.strip():
            e_class_single = f"class:{self.edge_class}, "

        return e_class_multi, e_class_single


class MatchNode:
    """Class to construct the node portions of a MATCH query."""

    def __init__(self):
        """Init method."""
        self.name: Optional[str] = None
        self.position: Optional[Position] = None
        self.node_class: Optional[str] = None
        self.namespace: Optional[str] = None
        self.bel: Optional[str] = None
        self.gene_path: bool = False
        self.how_name = None
        self.how_bel = None

    def set_outside(
        self,
        position: Position,
        name: Optional[str] = None,
        node_class: Optional[str] = None,
        namespace: Optional[str] = None,
        bel: Optional[str] = None,
        how_name: Optional[str] = SearchType.EXACT.value,
        how_bel: Optional[str] = SearchType.EXACT.value,
    ):
        """Assign attributes to the "outside" position."""
        self.position = position
        if name:
            self.name = name
        if node_class:
            self.node_class = node_class
        if namespace:
            self.namespace = namespace
        if bel:
            self.bel = bel
        if how_name:
            self.how_name = how_name if SearchType.has_value(how_name) else SearchType.EXACT.value
            print("how_name:", how_name, self.how_name)
        if how_bel:
            self.how_bel = how_bel if SearchType.has_value(how_bel) else SearchType.EXACT.value
            print("how_bel:", how_bel, self.how_bel)

    def set_inside(self, gene_path: bool, node_class: None):
        """Assign attributes to the "inside" position."""
        self.gene_path = gene_path
        self.position = Position.INSIDE
        self.node_class = node_class

    def get_node(self, alias_number: int) -> str:
        """Return node based on alias number."""
        print(f"hows: \n\tposition{self.position} \n\thow_name: {self.how_name}, \n\thow_bel: {self.how_bel}")
        namespace = f"namespace='{self.namespace}'"
        name = _get_where_by_how("name", self.name, self.how_name)
        bel = _get_where_by_how("bel", self.bel, self.how_bel)
        name_involved = f"('{self.name}' in involved_genes OR '{self.name}' in involved_other)"
        involved_genes = "involved_genes.size()>0"
        not_like_node = "$matched.n{}!=$currentMatch"
        where_inside_list = []

        where_str = ""

        if self.position in (Position.FIRST, Position.LAST):
            where_first_last = []
            if self.bel:
                where_first_last.append(bel)
            if self.namespace and not self.name:
                where_first_last.append(namespace)
            elif self.name and not self.namespace:
                if self.node_class in ["gene", "rna", "protein"]:
                    where_first_last.append(name)
                else:
                    where_first_last.append(name_involved)
            elif self.name and self.namespace:
                where_first_last.append(f"{name} AND {namespace}")
            where_str = " AND ".join(where_first_last)

        if self.position == Position.LAST:
            where_inside_list.append(not_like_node.format(1))
        if self.gene_path and not any([self.node_class, self.name, self.namespace]):
            where_inside_list.append(involved_genes)
        where_str = " AND ".join([x for x in ([where_str] + where_inside_list) if x])

        alias = f"as:n{alias_number}"
        where = f"where:({where_str})" if where_str else ""
        node_class = self.get_node_class()
        node_query_str = ", ".join([x for x in [node_class, where, alias] if x])

        return "{" + node_query_str + "}"

    def get_node_class(self):
        """Return node class of match query."""
        node_class = ""
        if self.node_class:
            node_class = f"class:{self.node_class}"
        return node_class


class GraphType(EnumExtension):
    """Not sure why there is a class for defining constants."""

    NODES = "nodes"
    EDGES = "edges"


class PathsResult(NamedTuple):
    """Class to build results of path query."""

    edge_paths_by_length: EdgePathsByLength
    unique_edges: Dict[Rid, EdgeInfo]
    unique_nodes: Dict[Rid, NodeInfo]


class PathQuery:
    """Class for constructing a path-based query."""

    def __init__(
        self,
        start_name: str,
        end_name: str,
        min_length: int,
        max_length: int,
        start_how_name: Optional[str] = None,
        end_how_name: Optional[str] = None,
        start_class: Optional[str] = None,
        end_class: Optional[str] = None,
        start_ns: Optional[str] = None,
        end_ns: Optional[str] = None,
        start_bel: Optional[str] = None,
        start_how_bel: Optional[str] = None,
        end_bel: Optional[str] = None,
        end_how_bel: Optional[str] = None,
        gene_path: bool = False,
        edge_class: Optional[str] = None,
        multiple_edge_classes: Optional[str] = None,
        inside_node_class: Optional[str] = None,
        mesh_term: Optional[str] = None,
        pmids: str = "",
        belish: Optional[str] = None,
        limit: int = 0,
        skip: int = 0,
    ):
        """Init method."""
        self.multiple_edge_classes = multiple_edge_classes
        self.limit = limit

        if isinstance(self.limit, str) and self.limit.isnumeric():
            self.limit = int(self.limit)

        self.skip = skip

        if isinstance(self.skip, str) and self.skip.isnumeric():
            self.skip = int(self.skip)

        self.pmids = [int(x.strip()) for x in pmids.split(",") if x.strip().isdigit()]
        self.mesh_terms = [x.strip() for x in mesh_term.split(";") if x.strip()]
        self.execute = Bel().execute
        self.min_length = min_length
        self.max_length = max_length
        self.edge_class = edge_class
        self.belish = belish
        self.max_paths = 100000
        self.max_unique_edges = 1000
        self.nodes = [MatchNode() for _ in range(self.max_length + 1)]

        self.nodes[0].set_outside(
            position=Position.FIRST,
            name=start_name,
            node_class=start_class,
            namespace=start_ns,
            bel=start_bel,
            how_name=start_how_name,
            how_bel=start_how_bel,
        )

        self.nodes[-1].set_outside(
            position=Position.LAST,
            name=end_name,
            node_class=end_class,
            namespace=end_ns,
            bel=end_bel,
            how_name=end_how_name,
            how_bel=end_how_bel,
        )

        self.edges = [
            MatchEdge(self.edge_class, self.multiple_edge_classes, self.mesh_terms, self.pmids)
            for _ in range(self.max_length)
        ]
        self.edges[-1].set_last(end_class)

        # TODO: why not using normal edge_class?
        for node in self.nodes[1:-1]:
            node.set_inside(gene_path, inside_node_class)

        self.too_many_paths = (
            "With the path length of {} we found already more than "
            f"{self.max_paths} pathways. Please specify you query (or set limit) and run again."
        )

        self.too_many_edges = (
            f"We found too many unique edges ({{}}, max allowed={self.max_unique_edges} ) with an "
            f"allowed maximum of {self.max_paths} paths. Please specify you query and run again. "
            "Decrease max path length, use limit or state start- and end-node more precisely."
        )

    def get_query_str(self, number_of_edges):
        """Create query string by number of edges."""
        query = "match " + self.nodes[0].get_node(1)
        edges = self.edges[-1 * number_of_edges :]
        nodes = self.nodes[-1 * (number_of_edges + 1) :]
        for i in range(1, number_of_edges + 1):
            query += edges[i - 1].get_edge(alias_number=i) + nodes[i].get_node(alias_number=i + 1)

        query += self.get_match_return(number_of_edges)
        return query

    @staticmethod
    def _get_unique_rids(graph_type: GraphType, path_length_dict: PathLengthDict) -> Set[Rid]:
        """Get unique node or edge rid set."""
        rids = {
            w
            for z in [[x for y in [en[graph_type.value] for en in v] for x in y] for _, v in path_length_dict.items()]
            for w in z
        }
        return rids

    def get_unique_edge_list(self, path_length_dict: PathLengthDict) -> Dict[Rid, EdgeInfo]:
        """Get unique list of edges."""
        edge_rids = self._get_unique_rids(GraphType.EDGES, path_length_dict)
        return {rid: self.get_edge_info(rid) for rid in edge_rids}

    def get_unique_node_list(self, path_length_dict: PathLengthDict) -> Dict[Rid, NodeInfo]:
        """Get unique list of nodes."""
        node_rids = self._get_unique_rids(GraphType.NODES, path_length_dict)
        return {rid: self.get_node_info(rid) for rid in node_rids}

    def get_edge_info(self, rid: Rid):
        """Get edge metadata by given rID."""
        sql = (
            f"Select out.@rid.asString() as subject_rid, in.@rid.asString() as object_rid, "
            f"out.bel as subject_bel, in.bel as object_bel,"
            f"@class.asString() as class, citation, evidence, pmid, annotation.mesh from {rid}"
        )
        return self.execute(sql)[0].oRecordData

    def get_node_info(self, rid: Rid):
        """Get node metadata by given rID."""
        sql = f"Select @class.asString() as class, * from {rid}"
        data = self.execute(sql)[0].oRecordData
        serializable_columns = get_columns(data["class"], exclude_non_serializable=True) + ["class"]
        serializable_data = {k: v for k, v in data.items() if k in serializable_columns}
        return serializable_data

    @property
    def allowed_edges(self) -> List[str]:
        """Return the number of allowed edges, includes both eBEL and BEL edge types."""
        bel_relations = list(get_bel_relation_types().keys())
        ebel_relations = list(get_ebel_relation_types().keys())
        return list(bel_relations + ebel_relations)

    def get_match_return(self, number_of_edges):
        """Standard way to return node and edges."""
        edges = [f"e{i}.@rid.asString()" for i in range(1, number_of_edges + 1)]
        edges_join = ",".join(edges)
        nodes = [f"n{i}.@rid.asString()" for i in range(1, number_of_edges + 2)]
        nodes_join = ",".join(nodes)

        if self.limit and self.limit <= (self.max_paths + 1):
            limit = self.limit
        else:
            limit = self.max_paths + 1

        skip = f" skip {self.skip} " if self.skip else ""

        return f" return [{edges_join}] as edges, [{nodes_join}] as nodes {skip} limit {limit}"

    def get_query_str_belish_num_edges(self):
        """Find the number of edges that match the BELish query string."""
        node_strings, edges = [], []

        if self.belish:
            node_strings, edges = self.get_belish_nodes_edges()

        if len(node_strings) == len(edges) + 1:
            match_str = self.get_belish_match_str(edges, node_strings)

            return match_str, len(edges)

    def get_belish_match_str(self, edges, node_strings):
        """Build the MATCH string based on the BELish query."""
        # ALERT: if a where in the following node multi class of edge are ignored
        edge_direction = {
            "->": {
                "one_class": ".outE(){{class:{},as:e{}{}}}.inV()",
                "multi_class": ".outE({}){{as:e{}{}}}.inV()",
            },
            "<-": {
                "one_class": ".inE(){{class:{},as:e{}{}}}.outV()",
                "multi_class": ".inE({}){{as:e{}{}}}.outV()",
            },
        }
        re_node_in_box = re.compile(
            r'^\[\s*(?P<class_name>\w+)(?P<params>(\s+\w+(\.\w+)?(!=|=|>|<|~|\*)(\d+|\d+\.\d+|[\w%]+|"[^"]+"))*)\s*\]$'
        )

        match_str = "match "
        for i in range(len(node_strings)):
            found_node_in_box = re_node_in_box.search(node_strings[i])
            if found_node_in_box:
                node_where = ""
                node_groups = found_node_in_box.groupdict()
                if node_groups["params"]:
                    node_where = self.get_where_list_by_params(node_groups["params"])
                match_str += f"{{class:{node_groups['class_name']} {node_where}, as:n{i + 1}}}"
            else:
                match_str += f"{{class:bel, where:(bel like '{node_strings[i]}'), as:n{i + 1}}}"
            if i <= len(edges) - 1:
                edge_temp = edge_direction[edges[i].direction]
                edge_class_names = [x.strip() for x in edges[i].name.split(",") if x.strip()]
                edge_where = ""
                if edges[i].params_str:
                    edge_where = self.get_where_list_by_params(edges[i].params_str)
                if len(edge_class_names) == 1:
                    match_str += edge_temp["one_class"].format(edge_class_names[0], i + 1, edge_where)
                else:
                    edge_class_names_joined = ",".join([f'"{x}"' for x in edge_class_names])
                    match_str += edge_temp["multi_class"].format(edge_class_names_joined, i + 1, edge_where)

        match_str += self.get_match_return(len(edges))
        return match_str

    @staticmethod
    def get_where_list_by_params(params_str):
        """Build WHERE section of query based on the passed parameters."""
        where_list = []
        re_params_in_box = re.compile(r'(\w+(\.\w+)?)(!=|=|>|<|~|\*)(\d+|\d+\.\d+|[\w%]+|"[^"]+")')
        for param, sub_param, operator, value in re_params_in_box.findall(params_str):
            operator = "like" if operator == "~" else operator
            operator = "in" if operator == "*" else operator
            equals_or_in_and_number = operator in ["=", "in"] and re.search(r"^\d+(\.\d+)?$", value)
            quotes_surrounded = re.search('^".*"$', value)
            if not (operator in [">", "<"] or equals_or_in_and_number or quotes_surrounded):
                value = f'"{value}"'
            if operator == "in":
                where_list.append(f"{value} {operator} {param}")
            else:
                where_list.append(f"{param} {operator} {value}")
        where = ", where:(" + " AND ".join(where_list) + ")"
        return where

    def get_belish_nodes_edges(self) -> Tuple[List[str], List[BELishEdge]]:
        """Return all BELish nodes and edges."""
        r = re.split(
            r"\s+(-(([a-z_0-9,]+)(\s+.*?)?)(->)|(<-)(([a-z_0-9,]+)(\s+.*?)?)-|-(->)|(<-)-)\s+",
            self.belish,
        )
        nodes: List[str] = r[::12]
        edge_zip = zip(
            r[8::12],
            r[3::12],
            r[1::12],
            r[5::12],
            r[6::12],
            r[10::12],
            r[11::12],
            r[4::12],
        )
        edges: List[BELishEdge] = [
            BELishEdge(x[0] or x[1] or "", x[3] or x[4] or x[5] or x[6], x[7]) for x in edge_zip
        ]
        return nodes, edges

    def get_paths(self) -> Union[PathsResult, Dict]:
        """Get paths by query."""
        self.max_paths = 100000
        if self.edge_class and not (self.edge_class in self.allowed_edges or self.edge_class == "E"):
            return {"error": "Unknown relation type."}

        path_length_dict: PathLengthDict = {}
        edge_paths_by_length: EdgePathsByLength = {}

        for number_of_edges in range(self.min_length, self.max_length + 1):
            query_str = self.get_query_str(number_of_edges)
            print(query_str)
            paths: List[Dict[str, Any]] = [x.oRecordData for x in self.execute(query_str)]

            if len(paths) > self.max_paths:
                return {"error": self.too_many_paths.format(number_of_edges)}

            path_length_dict[number_of_edges] = paths
            edge_paths_by_length[number_of_edges] = [x["edges"] for x in paths]

        unique_edges: Dict[Rid, EdgeInfo] = self.get_unique_edge_list(path_length_dict)
        if len(unique_edges) > self.max_unique_edges:
            return {"error": self.too_many_edges.format(len(unique_edges))}

        unique_nodes: Dict[Rid, NodeInfo] = self.get_unique_node_list(path_length_dict)
        paths_results = PathsResult(
            edge_paths_by_length=edge_paths_by_length,
            unique_edges=unique_edges,
            unique_nodes=unique_nodes,
        )
        return paths_results

    def get_paths_by_belish(self):
        """Get paths by BELish query."""
        self.max_paths = 100
        path_length_dict: PathLengthDict = {}
        edge_paths_by_length: EdgePathsByLength = {}
        query_str_belish_num_edges = self.get_query_str_belish_num_edges()
        print(query_str_belish_num_edges)

        if query_str_belish_num_edges:
            query_str, number_of_edges = query_str_belish_num_edges
            paths: List[Dict[str, Any]] = [x.oRecordData for x in self.execute(query_str)]
            if len(paths) > self.max_paths:
                return {"error": self.too_many_paths.format(number_of_edges)}

            path_length_dict[number_of_edges] = paths
            edge_paths_by_length[number_of_edges] = [x["edges"] for x in paths]
            unique_edges: Dict[Rid, EdgeInfo] = self.get_unique_edge_list(path_length_dict)
            if len(unique_edges) > self.max_unique_edges:
                return {"error": self.too_many_edges.format(len(unique_edges))}

            unique_nodes: Dict[Rid, NodeInfo] = self.get_unique_node_list(path_length_dict)
            paths_results = PathsResult(
                edge_paths_by_length=edge_paths_by_length,
                unique_edges=unique_edges,
                unique_nodes=unique_nodes,
            )
            return paths_results


def _get_number(number_string: str, default_value: int, min_value: int = None, max_value: int = None) -> int:
    """Parse the number string.

    Check if the number_string is numeric and >= min_value or <= max_value, otherwise assign min_value
    respectively max_value.
    """
    return_value: Optional[int] = None
    if number_string is not None and number_string.isnumeric():
        number = int(number_string)
        if min_value:
            return_value = number if number >= min_value else min_value
        elif max_value:
            return_value = number if number <= max_value else max_value
    else:
        if min_value:
            return_value = default_value
        elif max_value:
            return_value = default_value
    return return_value


def _get_path_query() -> Union[PathQuery, ErrorMessage]:
    """Return paths found for query.

    Raises
    ------
    ErrorMessage
    """
    start_name = request.args.get("start_node_name")
    end_name = request.args.get("end_node_name")
    start_how_name = request.args.get("start_how_node_name")
    end_how_name = request.args.get("end_how_node_name")
    start_class = request.args.get("start_node_class", "bel")
    end_class = request.args.get("end_node_class", "bel")
    start_ns = request.args.get("start_node_namespace")
    end_ns = request.args.get("end_node_namespace")
    start_bel = request.args.get("start_bel")
    start_how_bel = request.args.get("start_how_bel")
    end_bel = request.args.get("end_bel")
    end_how_bel = request.args.get("end_how_bel")
    edge_class = request.args.get("connecting_relation")
    multiple_edge_classes = request.args.get("multiple_connecting_relations", "")
    inside_node_class = request.args.get("connecting_node_class")
    gene_path = request.args.get("only_gene_related_nodes_on_path")
    pmid = request.args.get("pmid", "")
    mesh_term = request.args.get("mesh_term", "")
    gene_path = True if gene_path == "true" else False
    limit = request.args.get("limit", "")
    limit = int(limit) if limit.isnumeric() else 0
    belish = request.args.get("belish", "")

    min_value = 1
    max_value = 10
    min_str = request.args.get("min_path_length")
    min_length = _get_number(min_str, 1, min_value=min_value)
    max_str = request.args.get("max_path_length")
    max_length = _get_number(max_str, 3, max_value=max_value)

    if pmid:
        edge_class = "bel_relation"
        end_class = "bel"
        max_length = 1

    path_query = PathQuery(
        start_name=start_name,
        end_name=end_name,
        min_length=min_length,
        max_length=max_length,
        start_how_name=start_how_name,
        end_how_name=end_how_name,
        start_class=start_class,
        end_class=end_class,
        start_ns=start_ns,
        end_ns=end_ns,
        start_bel=start_bel,
        start_how_bel=start_how_bel,
        end_bel=end_bel,
        end_how_bel=end_how_bel,
        gene_path=gene_path,
        edge_class=edge_class,
        multiple_edge_classes=multiple_edge_classes,
        inside_node_class=inside_node_class,
        mesh_term=mesh_term,
        pmids=pmid,
        belish=belish,
        limit=limit,
    )
    return path_query


def get_paths() -> Union[dict, PathQuery]:
    """Return paths found for query."""
    print("get_paths:\n\n", request.args)
    path_query = _get_path_query()
    if isinstance(path_query, PathQuery):
        paths = path_query.get_paths()
        if isinstance(paths, dict):
            return paths
        else:
            return paths._asdict()
    else:
        return path_query


def get_paths_by_belish() -> Union[dict, PathQuery]:
    """Find all paths from given BELish query and return results as dictionary."""
    path_query = _get_path_query()
    if isinstance(path_query, PathQuery):
        paths_by_belish = path_query.get_paths_by_belish()
        print("path_query:", type(paths_by_belish))
        if isinstance(paths_by_belish, dict):
            return paths_by_belish
        else:
            return paths_by_belish._asdict()
    else:
        return path_query


def get_paths_as_dot():
    """Execute protected method."""
    path_query = _get_path_query()
    return _get_paths_as_dot(path_query.get_paths())


def get_paths_by_belish_as_dot():
    """Find all paths from given BELish query and return results as DOT."""
    path_query = _get_path_query()
    return _get_paths_as_dot(path_query.get_paths_by_belish())


def _get_paths_as_dot(paths):
    """Find all paths between specified query and return results as DOT."""
    if isinstance(paths, PathsResult):
        edges = defaultdict(int)
        d = Digraph()
        d.attr("graph", fontname="helvetica")
        d.attr("node", shape="note")
        if len(paths.unique_nodes) == 0:
            row_template = '<B><FONT POINT-SIZE="6">{}:</FONT></B>' '<FONT POINT-SIZE="6">{}</FONT>'
            key_value_dict = {}
            for k, v in request.args.items():
                key_value_dict[k] = cgi.html.escape(v).encode("ascii", "xmlcharrefreplace").decode("utf-8")
            legend_rows = "<BR/>".join([row_template.format(k, v) for k, v in key_value_dict.items()])
            d.node(
                "legend",
                f'<<FONT POINT-SIZE="16">NO PATHS FOUND!</FONT><BR/>{legend_rows}>',
            )

        d.attr("node", shape="box")
        d.attr("node", style="filled")

        for rid, v in paths.unique_nodes.items():
            node_id = rid.replace(":", ".")
            d.attr("node", fillcolor=node_colours.get(v["class"], "grey"))
            view_labels = get_node_view_labels(v["class"])

            sub_label = view_labels["sub_label"]
            if "involved_genes" in v or "involved_other" in v:
                involved = (
                    ",".join(v["involved_genes"] + v["involved_other"]).replace("<", "&lt;").replace(">", "&gt;")
                )
                bel_str = v["bel"].replace("<", "&lt;").replace(">", "&gt;")
                node_label = (
                    f'<<FONT POINT-SIZE="10">{v["class"]}</FONT><BR/><FONT POINT-SIZE="16">{involved}</FONT>'
                    f'<BR/><FONT POINT-SIZE="6">{bel_str}</FONT>>'
                )
            else:
                label_col_list = view_labels["label"]
                label_value = ""
                if len(label_col_list) == 1:
                    label_value = v[label_col_list[0]]
                elif len(label_col_list) >= 1:
                    label_value = "; ".join([str(v.get(x, "")) for x in label_col_list])

                sub_label_value = ""
                if sub_label and v.get(sub_label[0]):
                    sub_label_value = '</FONT><BR/><FONT POINT-SIZE="6">' + v[sub_label[0]]
                node_label = (
                    f'<<FONT POINT-SIZE="10">{v["class"]}</FONT><BR/>'
                    f'<FONT POINT-SIZE="16">{label_value}{sub_label_value}</FONT>>'
                )

            d.node(node_id, node_label)

        for v in paths.unique_edges.values():
            s_rid = v["subject_rid"].replace(":", ".")
            o_rid = v["object_rid"].replace(":", ".")
            edges[(s_rid, o_rid, v["class"])] += 1

        for edge, number_of_edge in edges.items():
            s_rid, o_rid, label = edge
            d.attr("edge", color=edge_colours.get(label, "grey"))
            d.edge(
                s_rid,
                o_rid,
                f'<<FONT POINT-SIZE="8">{label} [{number_of_edge}]</FONT>>',
            )
        return d.source
    else:
        return paths


def get_publication_year_statistics():
    """Return publication counts by year derived from BEL edges."""
    sql = (
        "Select year, count(*) as number_of_edges from (Select citation.pub_date.left(4) as year "
        "from bel_relation) where year!='' group by year order by year desc"
    )
    return [x.oRecordData for x in Bel().execute(sql)]


def get_class_infos():
    """Return node or edge class metadata."""
    b = Bel()
    sql = (
        "SELECT name, superClass, abstract, properties FROM (select expand(classes) "
        "FROM metadata:schema) WHERE NOT (name LIKE 'O%' OR name like '_%')"
    )
    class_name_dict = {}
    parent_dict = defaultdict(list)
    in_out_dict = {}
    in_edge_class_dict = defaultdict(list)
    out_edge_class_dict = defaultdict(list)
    for row in b.execute(sql):
        r = dict(row.oRecordData)
        in_out = {
            p["name"]: p["linkedClass"] for p in r["properties"] if "linkedClass" in p and p["name"] in ["in", "out"]
        }
        if in_out:
            in_out_dict[r["name"]] = in_out
            in_edge_class_dict[in_out["in"]].append(r["name"])
            out_edge_class_dict[in_out["out"]].append(r["name"])
        r.pop("properties")
        class_name_dict[r["name"]] = r
        if r.get("superClass"):  # all except roots
            parent_dict[r["superClass"]].append({"name": r["name"], "abstract": r["abstract"]})

    results = {}

    for class_name in class_name_dict:
        cnd = class_name_dict[class_name]
        result = {
            "abstract": cnd["abstract"],
            "parents_path": [class_name],
            "children": parent_dict[class_name],
        }
        check4parent = True
        while check4parent:
            last_parent = result["parents_path"][-1]
            if last_parent not in class_name_dict:
                break
            parent = class_name_dict[last_parent].get("superClass")
            if parent:
                result["parents_path"].append(parent)
            else:
                check4parent = False
        results[class_name] = result

        # get in_out from parents
        for parent in result["parents_path"]:
            if parent in in_out_dict:
                results[class_name]["in_out"] = in_out_dict[parent]
                break

        if class_name in in_edge_class_dict:
            results[class_name]["in_relations"] = in_edge_class_dict[class_name]
        if class_name in out_edge_class_dict:
            results[class_name]["out_relations"] = out_edge_class_dict[class_name]
    return results


def get_class_info_by_name():
    """Return node type by given name."""
    results = get_class_infos()
    return results.get(request.args.get("name"))


def get_class_infos_by_parent_name(childs_of) -> dict:
    """Get node or edge class information as DOT."""
    results = get_class_infos()
    return {k: v for k, v in results.items() if childs_of in v["parents_path"]}


def _get_class_info_as_dot(get_class_method):
    """Get node or edge class information as DOT."""
    classes = get_class_method()
    graph = Digraph()
    graph.graph_attr["rankdir"] = "LR"
    graph.node_attr["shape"] = "plaintext"
    for node_name, v in classes.items():
        for child in [x["name"] for x in v["children"]]:
            graph.edge(node_name, child)
    return graph.source


def get_bel_node_types():
    """Return BEL nodes and their metadata."""
    return get_class_infos_by_parent_name("bel")


def get_all_node_types():
    """Return BEL nodes and their metadata."""
    bel_node_types = get_class_infos_by_parent_name("bel")
    bel_node_types.update(get_class_infos_by_parent_name("ebel"))
    return bel_node_types


def get_bel_node_types_as_dot():
    """Return BEL nodes as DOT."""
    return _get_class_info_as_dot(get_bel_node_types)


def get_ebel_node_types():
    """Return eBEL added nodes and their metadata."""
    return get_class_infos_by_parent_name("ebel")


def get_ebel_node_types_as_dot():
    """Return eBEL added nodes as DOT."""
    return _get_class_info_as_dot(get_ebel_node_types)


def get_bel_relation_types():
    """Return BEL edges and their metadata."""
    return get_class_infos_by_parent_name("bel_relation")


def get_bel_relation_types_as_dot():
    """Return BEL edges as DOT."""
    return _get_class_info_as_dot(get_bel_relation_types)


def get_ebel_relation_types():
    """Return eBEL added edges and their metadata."""
    return get_class_infos_by_parent_name("ebel_relation")


def get_ebel_relation_types_as_dot():
    """Return eBEL added edges as DOT."""
    return _get_class_info_as_dot(get_ebel_relation_types)


# ApiResult = namedtuple('ApiResult', ['results', 'number_of_results', 'page', 'pages','page_size'])


def get_documents():
    """Return a list of documents that were imported to compile BEL graph."""
    sql = """Select
        @rid.asString() as rid,
        description,
        contact_info,
        version,
        licence,
        date.uploaded as uploaded,
        copyright,
        keywords.label as keywords,
        file.last_modified as file_last_modified,
        name as file_name,
        git_info.origin_url as git_origin_url,
        git_info.hexsha as git_hexsha,
        git_info.repo_path as git_repo_path,
        authors
    from bel_document"""
    results = [x.oRecordData for x in Bel().execute(sql)]
    len_results = len(results)
    return {
        "results": results,
        "number_of_results": len_results,
        "page": 1,
        "pages": 1,
        "page_size": len_results,
    }


def get_pmids():
    """Return all PMIDs and their counts from BEL edges."""
    sql = (
        "Select pmid, count(*) as number_of_edges, citation"
        " from bel_relation "
        "where pmid!=0 group by pmid order by number_of_edges desc"
    )
    return [x.oRecordData for x in Bel().execute(sql)]
